#include "init.h"
#include "bitboard.h"
#include "board.h"
#include <iostream>
#include <string.h>

using namespace std;

unsigned char first_ones[65536];
unsigned char last_ones[65536];

BITBOARD white_scastle_mask, white_lcastle_mask, black_scastle_mask, black_lcastle_mask;

void InitZeroMasks(){
	int i,j,mask;

	for (i = 0; i < 65536; i++){
		first_ones[i] = last_ones[i] = 0;
		for (j = 0, mask = 1; j < 16; j++, mask <<= 1){
			if (i & mask){
				first_ones[i] = j;
				break;
			}
		}
		for (j = 15, mask = 1 << 15; j >= 0 ; j--, mask >>= 1){
			if (i & mask){
				last_ones[i] = j;
				break;
			}
		}
	}
}



BITBOARD clear_mask_90[64];
BITBOARD clear_mask_a1[64];
BITBOARD clear_mask_h1[64];

BITBOARD set_mask_90[64];
BITBOARD set_mask_a1[64];
BITBOARD set_mask_h1[64];


static const unsigned char index_90[] = {
	7,15,23,31,39,47,55,63,
	6,14,22,30,38,46,54,62,
	5,13,21,29,37,45,53,61,
	4,12,20,28,36,44,52,60,
	3,11,19,27,35,43,51,59,
	2,10,18,26,34,42,50,58,
	1, 9,17,25,33,41,49,57,
	0, 8,16,24,32,40,48,56,
};

  /** Index for board rotation of 45 degrees left */
static const unsigned char index_a1[] = {
	28,36,43,49,54,58,61,63,
	21,29,37,44,50,55,59,62,
	15,22,30,38,45,51,56,60,
	10,16,23,31,39,46,52,57,
	 6,11,17,24,32,40,47,53,
	 3, 7,12,18,25,33,41,48,
	 1, 4, 8,13,19,26,34,42,
	 0, 2, 5, 9,14,20,27,35
};
		 
/** Index for board rotation of 45 degrees right*/
static const unsigned char index_h1[] = {
	63,62,60,57,53,48,42,35,
	61,59,56,52,47,41,34,27,
	58,55,51,46,40,33,26,20,
	54,50,45,39,32,25,19,14,
	49,44,38,31,24,18,13, 9,
	43,37,30,23,17,12, 8, 5,
	36,29,22,16,11, 7, 4, 2,
	28,21,15,10, 6, 3, 1, 0};


void InitBitboardMasks(){
	for (int i = 0; i < 64; i++){
		set_mask_90[i] = OneShl(index_90[i]);
		set_mask_a1[i] = OneShl(index_a1[i]);
		set_mask_h1[i] = OneShl(index_h1[i]);

		clear_mask_90[i] = Compl(set_mask_90[i]);
		clear_mask_a1[i] = Compl(set_mask_a1[i]);
		clear_mask_h1[i] = Compl(set_mask_h1[i]);
	}

	white_scastle_mask = Or(OneShl(F1),OneShl(G1));
	white_lcastle_mask = Or(Or(OneShl(B1),OneShl(C1)),OneShl(D1));
	black_scastle_mask = Or(OneShl(F8),OneShl(G8));
	black_lcastle_mask = Or(Or(OneShl(B8),OneShl(C8)),OneShl(D8));
}


BITBOARD rank_mask[8];
BITBOARD file_mask[8]; 

BITBOARD set_mask[64]; 

BITBOARD mask_left_edge;
BITBOARD mask_right_edge;

BITBOARD mask_rank567;
BITBOARD mask_rank234;


void InitRankFileMasks(){
	int i;

	rank_mask[RANK1] = 0xff;

	for (i = RANK2; i <= RANK8; i++)
		rank_mask[i] = Shl(rank_mask[i-1],8);

	file_mask[FILEA] = Compl((BITBOARD)-72340172838076674);


	for (i = FILEB; i <= FILEH; i++)
		file_mask[i] = Shl(file_mask[i-1],1);

	mask_left_edge = Compl(file_mask[FILEA]);
	mask_right_edge = Compl(file_mask[FILEH]);

	for (i = 0; i < 64; i++)
		set_mask[i] = OneShl(i);

	mask_rank567 = rank_mask[RANK5] | rank_mask[RANK6] | rank_mask[RANK7];
	mask_rank234 = rank_mask[RANK2] | rank_mask[RANK3] | rank_mask[RANK4];

	/*
	for (i = 0; i < 8; i++){
		cout << "rank mask " << i << endl;
		PrintBitboard(rank_mask[i]);
	}
	for (i = 0; i < 8; i++){
		cout << "file mask " << i << endl;
		PrintBitboard(file_mask[i]);
	}
	*/
}


int FillBits(int x,int start) {
	int i;
		
	int y = 0;
		
	for (i = start - 1; (i>= 0) && ((x & (1 << i)) == 0); i--)
		y |= (int)(1 << i);
	if (i >= 0)
		y |= (int)(1 << i);

	for (i = start + 1; (i<= 7) && ((x & (1 << i)) == 0); i++)
		y |= (int)(1 << i);
	if (i <= 7)
		y |= (int)(1 << i);
			
	return y;
}


BITBOARD rook_attacks[64][256];
BITBOARD rook_attacks90[64][256];

DiagInfo bishop_attacks[64];

BITBOARD knight_attacks[64];
BITBOARD king_attacks[64];

BITBOARD wpawn_attacks[64];
BITBOARD bpawn_attacks[64];


static unsigned char index_file[] = {
	0, 8,16,24,32,40,48,56,
	1, 9,17,25,33,41,49,57,
	2,10,18,26,34,42,50,58,
	3,11,19,27,35,43,51,59,
	4,12,20,28,36,44,52,60,
	5,13,21,29,37,45,53,61,
	6,14,22,30,38,46,54,62,
	7,15,23,31,39,47,55,63,
};


/*
void PrintX(int x){
	for (int j = 7; j >= 0; j--)
		if (x & (1 << j))
			cout << '1';
		else
			cout << '0';
	cout << endl;
}
*/

inline int ByteReverse(unsigned char x){
	int result = 0;
	for (int i = 0; i < 8; i++){
		result <<= 1;
		if (x & (1 << i))
			result += 1;
	}
	return result;
}

void InitRookLut(){
	int i,j,k;
	BITBOARD file,x;

	for (k = 0; k < 256; k++){

		// normal lut	
		for (i = 0; i < 8; i++){
			rook_attacks[i][k] = FillBits(k, i);
		}

		for (j = 8; j < 64; j++)
			rook_attacks[j][k] = Shl(rook_attacks[j-8][k], 8);

		// 90 degrees rotated lut	

		for (i = 0; i < 8; i++){
			x = FillBits(ByteReverse(k), i);
			file = 0;
			for (j = 0; j < 8; j++){
				file = Shl(file, 8);
				if (And(x, OneShl(7 - j)))
					file += 1;
			}

			rook_attacks90[index_file[i]][k] = file;
		}

		for (j = 8; j < 64; j++)
			rook_attacks90[index_file[j]][k] = Shl(rook_attacks90[index_file[j-8]][k], 1);
	}
}

void InitKnightLut(){
	int i,j,p,rank,file;
	BITBOARD bb;

	int add_i[] = { -1,-1, 1, 1, -2, -2,  2, 2};
	int add_j[] = { -2, 2,-2, 2, -1,  1, -1, 1};

	for (i = 0; i < 8; i++)
		for (j = 0; j < 8; j++){
			bb = 0;

			for (p = 0; p < 8; p++){
				rank = i + add_i[p];
				file = j + add_j[p];
				if (rank >= 0 && file >= 0 && rank < 8 && file < 8)
					Set(bb, Square(rank,file));
			}

			knight_attacks[Square(i,j)] = bb;
		}
}

void InitKingLut(){
	int i,j,p,rank,file;
	BITBOARD bb;

	int add_i[] = { -1,-1,-1,  1, 1, 1,  0, 0};
	int add_j[] = { -1, 1, 0, -1, 1, 0, -1, 1};

	for (i = 0; i < 8; i++)
		for (j = 0; j < 8; j++){
			bb = 0;

			for (p = 0; p < 8; p++){
				rank = i + add_i[p];
				file = j + add_j[p];
				if (rank >= 0 && file >= 0 && rank < 8 && file < 8)
					Set(bb, Square(rank,file));
			}

			king_attacks[Square(i,j)] = bb;
		}
}


static const int diag_mask[] = {0,1,3,7,15,31,63,127,255};
	
static const unsigned char diag_a1_start[] = {
	0, 1, 2, 3, 4, 5, 6, 7,
	8, 0, 1, 2, 3, 4, 5, 6,
	16, 8, 0, 1, 2, 3, 4, 5,
	24,16, 8, 0, 1, 2, 3, 4,
	32,24,16, 8, 0, 1, 2, 3,
	40,32,24,16, 8, 0, 1, 2,
	48,40,32,24,16, 8, 0, 1,
	56,48,40,32,24,16, 8, 0
};
		 
static char diag_a1_pos[] = {
	0,0,0,0,0,0,0,0,
	0,1,1,1,1,1,1,1,
	0,1,2,2,2,2,2,2,
	0,1,2,3,3,3,3,3,
	0,1,2,3,4,4,4,4,
	0,1,2,3,4,5,5,5,
	0,1,2,3,4,5,6,6,
	0,1,2,3,4,5,6,7
};
		  
static char diag_a1_length[] = {
	8,7,6,5,4,3,2,1,
	7,8,7,6,5,4,3,2,
	6,7,8,7,6,5,4,3,
	5,6,7,8,7,6,5,4,
	4,5,6,7,8,7,6,5,
	3,4,5,6,7,8,7,6,
	2,3,4,5,6,7,8,7,
	1,2,3,4,5,6,7,8 
};


static char diag_h1_start[] = {
	0, 8,16,24,32,40,48,56,
	8,16,24,32,40,48,56,57,
	16,24,32,40,48,56,57,58,
	24,32,40,48,56,57,58,59,
	32,40,48,56,57,58,59,60,
	40,48,56,57,58,59,60,61,
	48,56,57,58,59,60,61,62,
	56,57,58,59,60,61,62,63 
};
			
static char diag_h1_pos[] = {
	0,1,2,3,4,5,6,7,
	0,1,2,3,4,5,6,6,
	0,1,2,3,4,5,5,5,
	0,1,2,3,4,4,4,4,
	0,1,2,3,3,3,3,3,
	0,1,2,2,2,2,2,2,
	0,1,1,1,1,1,1,1,
	0,0,0,0,0,0,0,0 
};
			
static char diag_h1_length[] = {
	1,2,3,4,5,6,7,8,
	2,3,4,5,6,7,8,7,
	3,4,5,6,7,8,7,6,
	4,5,6,7,8,7,6,5,
	5,6,7,8,7,6,5,4,
	6,7,8,7,6,5,4,3,
	7,8,7,6,5,4,3,2,
	8,7,6,5,4,3,2,1 
};

static const unsigned char diag_a1_shift[] = {
	28,36,43,49,54,58,61,63,
	21,28,36,43,49,54,58,61,
	15,21,28,36,43,49,54,58,
	10,15,21,28,36,43,49,54,
	6,10,15,21,28,36,43,49,
	3, 6,10,15,21,28,36,43,
	1, 3, 6,10,15,21,28,36,
	0, 1, 3, 6,10,15,21,28 
};
     
static const unsigned char diag_h1_shift[] = {
	63,61,58,54,49,43,36,28,
	61,58,54,49,43,36,28,21,
	58,54,49,43,36,28,21,15,
	54,49,43,36,28,21,15,10,
	49,43,36,28,21,15,10, 6,
	43,36,28,21,15,10, 6, 3,
	36,28,21,15,10, 6, 3, 1,
	28,21,15,10, 6, 3, 1, 0 
};


#include <iostream>

using namespace std;

void InitBishopLut(){
	int i,j,k;
	BITBOARD tmp, bb, x;

	for (k = 0; k < 256; k++)
		for (i = 0; i < 64; i++){

		DiagInfo& d = bishop_attacks[i];

		// generate for rotated right diagonal (a1-h8)
		x = FillBits(k, diag_a1_pos[i]) & diag_mask[diag_a1_length[i]];

		tmp = OneShl(diag_a1_start[i]);
		bb = 0;

		for (j = 0; j < diag_a1_length[i]; j++){
			if (x & (1 << j))
				bb |= tmp;
			tmp = Shl(tmp, 9);
		}

		d.a1[k] = bb;
		d.shift_a1 = diag_a1_shift[i];

		// generate for rotated left diagonal (h1-a8)
		x = FillBits(k, diag_h1_pos[i]) & diag_mask[diag_h1_length[i]];
		tmp = OneShl(diag_h1_start[i]);
		bb = 0;

		for (j = 0; j < diag_h1_length[i]; j++){
			if (And(x, OneShl(j)))
				bb |= tmp;
			tmp = Shr(tmp, 7);
		}

		d.h1[k] = bb;
		d.shift_h1 = diag_h1_shift[i];
	}
}

void InitPawnLut(){
	int i, file;
	BITBOARD wtmp, btmp;

	for (i = 0; i < 64; i++) {
		wtmp = btmp = 0;
    
		file = File(i);
    
		if (i <= 55) {
			if (file > 0) Set(wtmp,i+7);
			if (file < 7) Set(wtmp,i+9);
		}

		if (i >= 8) {
			if (file > 0) Set(btmp,i-9);
			if (file < 7) Set(btmp,i-7);
		}
    
		wpawn_attacks[i] = wtmp;
		bpawn_attacks[i] = btmp;
	}
}



/*

A 32 bit random number generator. An implementation in C of the algorithm given by
Knuth, the art of computer programming, vol. 2, pp. 26-27. We use e=32, so
we have to evaluate y(n) = y(n - 24) + y(n - 55) mod 2^32, which is implicitly
done by unsigned arithmetic.

*/

unsigned int Random32(){
	/*
	random numbers from Mathematica 2.0.
	SeedRandom = 1;
	Table[Random[Integer, {0, 2^32 - 1}]
	*/
	static unsigned long x[55] = {
		1410651636UL, 3012776752UL, 3497475623UL, 2892145026UL, 1571949714UL,
		3253082284UL, 3489895018UL, 387949491UL, 2597396737UL, 1981903553UL,
		3160251843UL, 129444464UL, 1851443344UL, 4156445905UL, 224604922UL,
		1455067070UL, 3953493484UL, 1460937157UL, 2528362617UL, 317430674UL,
		3229354360UL, 117491133UL, 832845075UL, 1961600170UL, 1321557429UL,
	    747750121UL, 545747446UL, 810476036UL, 503334515UL, 4088144633UL,
		2824216555UL, 3738252341UL, 3493754131UL, 3672533954UL, 29494241UL,
		1180928407UL, 4213624418UL, 33062851UL, 3221315737UL, 1145213552UL,
		2957984897UL, 4078668503UL, 2262661702UL, 65478801UL, 2527208841UL,
		1960622036UL, 315685891UL, 1196037864UL, 804614524UL, 1421733266UL,
		2017105031UL, 3882325900UL, 810735053UL, 384606609UL, 2393861397UL };
  
	static int init = 1;
	static unsigned long y[55];
	static int j, k;
	unsigned long ul;

	if (init){
		int i;

		init = 0;
		for (i = 0; i < 55; i++) y[i] = x[i];
		j = 24 - 1;
		k = 55 - 1;
	}

	ul = (y[k] += y[j]);
	if (--j < 0) j = 55 - 1;
	if (--k < 0) k = 55 - 1;
	return((unsigned int)ul);
}

BITBOARD Random64(){
	BITBOARD result,r1,r2;

	r1=Random32();
	r2=Random32();
	result= Or(r1,Shl(r2,32));
	return result;
}

HASHCODE RandomHash[2][7][64];	
HASHCODE white_hash, black_hash;	
HASHCODE whiteblack_hash;

HASHCODE pawn_hash[2][64];

void InitHashCodes(){
	int wtm, square;
	for (wtm = 0; wtm < 2; wtm++)
		for (square = 0; square < 64; square++)
			for (int piece = 0; piece < 7; piece++)
				RandomHash[wtm][piece][square] = Random64();

	white_hash = Random64();
	black_hash = Random64();
	whiteblack_hash = Xor(white_hash,black_hash);

	for (wtm = 0; wtm < 2; wtm++)
		for (square = 0; square < 64; square++)
			pawn_hash[wtm][square] = Random64();
}




int directions[64][64];
BITBOARD plus1dir[64],plus7dir[64],plus8dir[64],plus9dir[64];
BITBOARD minus1dir[64],minus7dir[64],minus8dir[64], minus9dir[64];

BITBOARD interpose_squares[64][64];


void InitPlusMinusDir(){
	int i,j,sq, lastsq;
	const int rook_sq[4] = {-8, -1, 1, 8};
	const int bishop_sq[4] = {-9, -7, 7, 9};

	BITBOARD tmp;

	memset(plus1dir,  0, 64 * sizeof(BITBOARD));
	memset(plus7dir,  0, 64 * sizeof(BITBOARD));
	memset(plus8dir,  0, 64 * sizeof(BITBOARD));
	memset(plus9dir,  0, 64 * sizeof(BITBOARD));
	memset(minus1dir, 0, 64 * sizeof(BITBOARD));
	memset(minus7dir, 0, 64 * sizeof(BITBOARD));
	memset(minus8dir, 0, 64 * sizeof(BITBOARD));
	memset(minus9dir, 0, 64 * sizeof(BITBOARD));
	memset(directions, 0, 64 * 64 * sizeof(int));

	for (i = 0; i < 64; i++)
		for (j = 0; j < 4; j++){
			lastsq = i;
			sq = i + bishop_sq[j];;
			while (	sq < 64 && 
					sq >= 0 && 
					Abs(lastsq/8-sq/8) == 1 && 
					Abs((lastsq&7)-(sq&7)) == 1){

				if (bishop_sq[j] == -9)	minus9dir[i] |= OneShl(sq);
				else if (bishop_sq[j] == -7) minus7dir[i] |= OneShl(sq);
				else if (bishop_sq[j] ==  7) plus7dir[i] |= OneShl(sq);
				else if (bishop_sq[j] ==  9) plus9dir[i] |= OneShl(sq);
				else Assert(false);

				lastsq = sq;
				sq += bishop_sq[j];
			}
		}

	for (i = 0; i < 64; i++)
		for (j = 0; j < 4; j++){
			lastsq = i;
			sq = i + rook_sq[j];;
			while (	sq < 64 && 
					sq >= 0 && 
					((Abs(lastsq/8-sq/8) == 0 && Abs((lastsq&7)-(sq&7)) == 1)||
					 (Abs(lastsq/8-sq/8) == 1 && Abs((lastsq&7)-(sq&7)) == 0))){

				if (rook_sq[j] == -8)	minus8dir[i] |= OneShl(sq);
				else if (rook_sq[j] == -1) minus1dir[i] |= OneShl(sq);
				else if (rook_sq[j] ==  1) plus1dir[i] |= OneShl(sq);
				else if (rook_sq[j] ==  8) plus8dir[i] |= OneShl(sq);
				else Assert(false);

				lastsq = sq;
				sq += rook_sq[j];
			}
		}

	for (i = 0; i < 64; i++){
		tmp = plus1dir[i];
		while (tmp){
			j = FirstOne(tmp);
			directions[i][j] = 1;
			Clear(tmp, j);
		}
		tmp = plus7dir[i];
		while (tmp){
			j = FirstOne(tmp);
			directions[i][j] = 7;
			Clear(tmp, j);
		}
		tmp = plus8dir[i];
		while (tmp){
			j = FirstOne(tmp);
			directions[i][j] = 8;
			Clear(tmp, j);
		}
		tmp = plus9dir[i];
		while (tmp){
			j = FirstOne(tmp);
			directions[i][j] = 9;
			Clear(tmp, j);
		}
		tmp = minus1dir[i];
		while (tmp){
			j = FirstOne(tmp);
			directions[i][j] = -1;
			Clear(tmp, j);
		}
		tmp = minus7dir[i];
		while (tmp){
			j = FirstOne(tmp);
			directions[i][j] = -7;
			Clear(tmp, j);
		}
		tmp = minus8dir[i];
		while (tmp){
			j = FirstOne(tmp);
			directions[i][j] = -8;
			Clear(tmp, j);
		}
		tmp = minus9dir[i];
		while (tmp){
			j = FirstOne(tmp);
			directions[i][j] = -9;
			Clear(tmp, j);
		}
	}


	for (i = 0; i < 64; i++)
		for (j = 0; j < 64; j++){
			tmp = 0;

			switch (directions[i][j]){
			case -9 : tmp = Xor(minus9dir[i],minus9dir[j]); break;
			case -8 : tmp = Xor(minus8dir[i],minus8dir[j]); break;
			case -7 : tmp = Xor(minus7dir[i],minus7dir[j]); break;
			case -1 : tmp = Xor(minus1dir[i],minus1dir[j]); break;
			case +1 : tmp = Xor(plus1dir[i],plus1dir[j]); break;
			case +7 : tmp = Xor(plus7dir[i],plus7dir[j]); break;
			case +8 : tmp = Xor(plus8dir[i],plus8dir[j]); break;
			case +9 : tmp = Xor(plus9dir[i],plus9dir[j]); break;
			default : break;
			}

			interpose_squares[i][j] = tmp & Compl(OneShl(i)) & Compl(OneShl(j));
		}
}


BITBOARD king_control[64]; // 2x2 squares around the king

void InitPawnControl(){
	for (int sq = 0; sq < 64; sq++){
		int rank = File(sq);
		int file = Rank(sq);

		BITBOARD value = 0;
		for (int i = -2; i <= 2; i++)
			for (int j = -2; j <= 2; j++)
				if (rank+i >= 0 && rank+i <= 7 &&
					file+j >= 0 && file+j <= 7 &&
					(i != 0 || j != 0))
						value |= OneShl(Square(file+j, rank+i));

		king_control[sq] = value;
	}

}


BITBOARD free_pawn_ctrl[2][64];
BITBOARD passed_pawn_mask[2];

BITBOARD knight_outpost_mask[2][64];

BITBOARD backward_pawn_mask[2][64];
BITBOARD isolated_pawn_mask[8];

void InitIsolatedPawnTables(){
	for (int i = 1; i <= 6; i++)
		isolated_pawn_mask[i] = file_mask[i-1] | file_mask[i+1];
	isolated_pawn_mask[0] = file_mask[1];
	isolated_pawn_mask[7] = file_mask[6];

	#if 0
	for (i = 0; i < 8; i++){
		cout << "mask[" << i << "]" << endl;
		PrintBitboard(isolated_pawn_mask[i]);
	}
	#endif
}


void InitBackwardPawnTables(){
	int sq, color, i;
	int f;

	register BITBOARD tmp;

	for (color = 0; color < 2; color++)
		for (sq = A2; sq <= H7; sq++){
			tmp = 0;
			f = File(sq);
			if (f > 0) tmp |= OneShl(sq-1);
			if (f < 7) tmp |= OneShl(sq+1);

			if (color == WHITE){
				for (i = Rank(sq)-1; i >= 1; i--){
					tmp = tmp | (tmp >> 8);
				}
			} else {
				for (i = Rank(sq)+1; i <= 6; i++){
					tmp = tmp | (tmp << 8);
				}
			}

			backward_pawn_mask[color][sq] = tmp;
		}

	#if 0
	cout << "White backward pawn tables : " << endl;
	for (sq = 0; sq < 64; sq++){
		cout << "Square[" << (char)(File(sq)+'a') << (char)(Rank(sq)+'1') << "]" << endl;
		PrintBitboard(backward_pawn_mask[WHITE][sq]);
	}
	
	cout << endl;
	cout << "Black backward pawn tables : " << endl;
	for (sq = 0; sq < 64; sq++){
		cout << "Square[" << (char)(File(sq)+'a') << (char)(Rank(sq)+'1') << "]" << endl;
		PrintBitboard(backward_pawn_mask[BLACK][sq]);
	}
	#endif
}


BITBOARD knight_danger_mask_kside[2], knight_danger_mask_qside[2];

BITBOARD passed_pawns_mask[2];

void InitPassedPawnTables(){
	int sq,color,i;
	int rank, file;
	BITBOARD tmp;
	
	passed_pawns_mask[WHITE] = rank_mask[RANK1]|rank_mask[RANK2]|rank_mask[RANK3]|rank_mask[RANK4];
	passed_pawns_mask[BLACK] = rank_mask[RANK5]|rank_mask[RANK6]|rank_mask[RANK7]|rank_mask[RANK8];
	
	memset(free_pawn_ctrl, 0, sizeof(free_pawn_ctrl));

	for (color = 0;  color < 2; color++)
		for (sq = A2; sq <= H7; sq++){
			tmp = 0;
			rank = Rank(sq);
			file = File(sq);
			
			if (rank >= RANK2 && rank <= RANK6 && color == WHITE){
				tmp |= OneShl(sq+8);
				if (file > FILEA) tmp |= OneShl(sq+7);
				if (file < FILEH) tmp |= OneShl(sq+9);
				for (i = rank+2; rank <= RANK6; rank++)
					tmp |= Shl(tmp, 8);
			}

			if (rank >= RANK2 && rank <= RANK7 && color == BLACK){
				tmp |= OneShl(sq-8);
				if (file > FILEA) tmp |= OneShl(sq-9);
				if (file < FILEH) tmp |= OneShl(sq-7);
				for (i = rank-2; rank >= RANK3; rank--)
					tmp |= Shr(tmp, 8);
			}

			knight_outpost_mask[color][sq] = tmp & Compl(file_mask[File(sq)]);

			if (color == WHITE)
				tmp = tmp & Compl(wpawn_attacks[sq]);
			else
				tmp = tmp & Compl(bpawn_attacks[sq]);

			free_pawn_ctrl[color][sq] = tmp;

		}


	tmp = 0;
	for (sq = A5; sq <= H7; sq++)
		tmp |= OneShl(sq);
	passed_pawn_mask[WHITE] = tmp;

	tmp = 0;
	for (sq = A2; sq <= H4; sq++)
		tmp |= OneShl(sq);
	passed_pawn_mask[BLACK] = tmp;

	#if 0
	cout << "White passed pawn tables : " << endl;
	for (sq = 0; sq < 64; sq++){
		cout << "Square[" << (char)(File(sq)+'a') << (char)(Rank(sq)+'1') << "]" << endl;
		PrintBitboard(free_pawn_ctrl[WHITE][sq]);
	}
	
	cout << endl;
	cout << "Black passed pawn tables : " << endl;
	for (sq = 0; sq < 64; sq++){
		cout << "Square[" << (char)(File(sq)+'a') << (char)(Rank(sq)+'1') << "]" << endl;
		PrintBitboard(free_pawn_ctrl[BLACK][sq]);
	}

	cout << "Passed pawn mask : " << endl;
	PrintBitboard(passed_pawn_mask[WHITE]);
	PrintBitboard(passed_pawn_mask[BLACK]);
	#endif
}


BITBOARD king_pawn_mask1[2][64];	// mask for pawns on rank = (king's rank + 2)
BITBOARD king_pawn_mask2[2][64];	// mask for pawns on rank > (king's rank + 2)

BITBOARD queenside_files_mask, kingside_files_mask, center_files_mask;


BITBOARD file_fgh_mask, file_abc_mask;

void InitKingMasks(){
	int sq,i;

	BITBOARD tmp;

	for (sq = 0; sq < 48; sq++){
		king_pawn_mask1[WHITE][sq] = OneShl(sq+16);
		
		tmp = 0;
		for (i = Rank(sq)+3; i <= 7; i++)
			tmp |= OneShl(i*8+File(sq));

		king_pawn_mask2[WHITE][sq] = tmp;
	}


	for (sq = 16; sq < 64; sq++){
		king_pawn_mask1[BLACK][sq] = OneShl(sq-16);
		
		tmp = 0;
		for (i = Rank(sq)-3; i >= 0; i--)
			tmp |= OneShl(i*8+File(sq));

		king_pawn_mask2[BLACK][sq] = tmp;
	}

	queenside_files_mask = file_mask[0] | file_mask[1] | file_mask[2];
	kingside_files_mask = file_mask[5] | file_mask[6] | file_mask[7];
	center_files_mask = file_mask[3] | file_mask[4];

	knight_danger_mask_kside[BLACK] = kingside_files_mask & (rank_mask[RANK5] | rank_mask[RANK6]);
	knight_danger_mask_qside[BLACK] = queenside_files_mask & (rank_mask[RANK5] | rank_mask[RANK6]);
	knight_danger_mask_kside[WHITE] = kingside_files_mask & (rank_mask[RANK3] | rank_mask[RANK4]);
	knight_danger_mask_qside[WHITE] = queenside_files_mask & (rank_mask[RANK3] | rank_mask[RANK4]);

	file_fgh_mask = file_mask[FILEF]|file_mask[FILEG]|file_mask[FILEH];
	file_abc_mask = file_mask[FILEA]|file_mask[FILEB]|file_mask[FILEC];

	#if 0
		PrintBitboard(queenside_files_mask);
		cout << endl;
		PrintBitboard(center_files_mask);
		cout << endl;
		PrintBitboard(kingside_files_mask);
		cout << endl;
		
		cout << "King pawn masks - WHITE : " << endl;

		for (sq = 0; sq < 64; sq++){
			cout << "Square[" << (char)(File(sq)+'a') << (char)(Rank(sq)+'1') << "]" << endl;
			PrintBitboard(king_pawn_mask1[WHITE][sq]);
			cout << "--" << endl;
			PrintBitboard(king_pawn_mask2[WHITE][sq]);
			cout << endl;
		}
		cout << "King pawn masks - BLACK : " << endl;
		for (sq = 0; sq < 64; sq++){
			cout << "Square[" << (char)(File(sq)+'a') << (char)(Rank(sq)+'1') << "]" << endl;
			PrintBitboard(king_pawn_mask1[BLACK][sq]);
			cout << "--" << endl;
			PrintBitboard(king_pawn_mask2[BLACK][sq]);
			cout << endl;
		}
	#endif
}

BITBOARD pawn_kingsquare[2][64];


void InitPawnRaceTables(){
	int sq, left, right, up, down;
	int i,j;
	BITBOARD tmp;

	memset(pawn_kingsquare, 0, sizeof(pawn_kingsquare));

	for (sq = A2; sq <= H7; sq++){
		down = Rank(sq);
		up = 8;

		left = right = File(sq);

		tmp = 0;
		for (i = sq+7; File(i) <= File(sq) && Rank(i) < 8; i+= 7)
			left--;
		for (i = sq+9; File(i) >= File(sq) && Rank(i) < 8; i+= 9)
			right++;

		for (i = down; i <= up; i++)
			for (j = left; j <= right; j++)
				tmp |= OneShl(i*8+j);

		pawn_kingsquare[WHITE][sq] = tmp;
		pawn_kingsquare[BLACK][File(sq) + (7-Rank(sq))*8] = Mirror(tmp);

	}

	// finishing touches : pawns on 2nd rank for white and 7th rank for black
	// can move 2 squares forward, so we take this into consideration
	for (sq = A2; sq <= H2; sq++)
		pawn_kingsquare[WHITE][sq] = pawn_kingsquare[WHITE][sq+8];
	for (sq = A7; sq <= H7; sq++)
		pawn_kingsquare[BLACK][sq] = pawn_kingsquare[BLACK][sq-8];


	#if 0

	cout << "King race masks - WHITE : " << endl;
	for (sq = 0; sq < 64; sq++){
		cout << "Square[" << (char)(File(sq)+'a') << (char)(Rank(sq)+'1') << "]" << endl;
		PrintBitboard(pawn_kingsquare[WHITE][sq]);
		cout << endl;
	}
	cout << "King race masks - BLACK : " << endl;
	for (sq = 0; sq < 64; sq++){
		cout << "Square[" << (char)(File(sq)+'a') << (char)(Rank(sq)+'1') << "]" << endl;
		PrintBitboard(pawn_kingsquare[BLACK][sq]);
		cout << endl;
	}

	#endif
}

int grouped_pawns_bonus[256];

const int pawn_group_bonus[6] = { 0, 50, 50, -50, -50, -600};

void InitGroupedPawnLUT(){
	for (int x = 0; x <= 255; x++){
		int n_groups = 0;
		if (x & 1) n_groups = 1;
		for (int i = 1; i < 8; i++){
			if ((x & (1 << i)) && ((x & (1 << (i-1))) == 0))
				n_groups++;
		}
		grouped_pawns_bonus[x] = pawn_group_bonus[n_groups];
	}
}


BITBOARD open_files[256];

void InitOpenFilesLUT(){
	for (int i = 0; i <= 255; i++){
		BITBOARD tmp = 0;
		for (int j = 0; j < 8; j++)
			if ((i & (1<<j)) == 0)
				tmp |= file_mask[j];
		open_files[i] = tmp;
	}
}


void InitTables(){
	InitZeroMasks();

	InitRankFileMasks();
	
	InitBitboardMasks();

	InitRookLut();
	InitBishopLut();
	InitPawnLut();
	InitKnightLut();
	InitKingLut();

	InitPawnControl();

	InitHashCodes();

	InitPlusMinusDir();

	InitPassedPawnTables();
	InitBackwardPawnTables();
	InitIsolatedPawnTables();

	InitKingMasks();

	InitPawnRaceTables();
	
	InitGroupedPawnLUT();

	InitOpenFilesLUT();
}
