#include "board.h"
#include "hash.h"
#include <iostream>
#include <fstream>
#include "stats.h"
#include <string.h>

using namespace std;

static int white_score, black_score;


extern int directions[64][64];
extern BITBOARD plus1dir[64],plus7dir[64],plus8dir[64],plus9dir[64];
extern BITBOARD minus1dir[64],minus7dir[64],minus8dir[64], minus9dir[64];

extern BITBOARD interpose_squares[64][64];


extern BITBOARD free_pawn_ctrl[2][64];
extern BITBOARD backward_pawn_mask[2][64];
extern BITBOARD isolated_pawn_mask[8];


PawnHashEntry pawn_score;
PawnHashEntry* pscore = NULL;
static int pawns_on_file[2];




#define DOUBLED_PAWN_PENALTY		85
#define ISOLATED_PAWN_PENALTY		100
#define BACKWARDS_PAWN_PENALTY		80
#define PASSED_PAWN_BONUS			100

//#define BLOCKED_BISHOP_PENALTY		100

extern BITBOARD passed_pawns_mask[2];

extern BITBOARD queenside_files_mask, kingside_files_mask, center_files_mask;


void ComputePawnMasks(){
	int i;

	pawns_on_file[WHITE] = pawns_on_file[BLACK] = 0;
	for (i = 7; i >= 0; i--){
		pawns_on_file[WHITE] <<= 1;
		if (board.wPawns & file_mask[i])
			pawns_on_file[WHITE] |= 1;
		pawns_on_file[BLACK] <<= 1;
		if (board.bPawns & file_mask[i])
			pawns_on_file[BLACK] |= 1;
	}

	//PrintBitboard(pawns_on_file[WHITE]); cout << endl;
	//PrintBitboard(pawns_on_file[BLACK]); cout << endl;
}


const BITBOARD blocked_rook_mask_queenside_w = OneShl(A1) | OneShl(B1) | OneShl(A2) | OneShl(A3);
const BITBOARD blocked_rook_mask_kingside_w  = OneShl(H1) | OneShl(G1) | OneShl(H2) | OneShl(H3);
const BITBOARD blocked_rook_mask_queenside_b = OneShl(A8) | OneShl(B8) | OneShl(A7) | OneShl(A6);
const BITBOARD blocked_rook_mask_kingside_b  = OneShl(H8) | OneShl(G8) | OneShl(H7) | OneShl(H6);


const int CONNECTED_ROOKS_BONUS = 60;

const int ROOK_ON_7TH_LINE_BONUS2_ONE_ROOK  = 250;
const int ROOK_ON_7TH_LINE_BONUS2_TWO_ROOKS = 300;

const int ROOK_MOBILITY_BONUS = 10;

//const int ROOK_ON_8TH_LINE_BONUS = 150;

const int ROOK_ON_OPEN_FILE_BONUS = 100;
const int ROOK_ON_SEMIOPEN_FILE_BONUS = 25;
//const int ROOK_ON_OPEN_FILE_ATTACKING_ISOLATED_PAWN = 200;

extern BITBOARD open_files[255];


const int rook_on_7th_bonus_onerook[8]  = { 800,  550, 500, 300, 150,  50,  50,  50};
const int rook_on_7th_bonus_tworooks[8] = {1000,  800, 700, 700, 500, 400, 400, 400};

const int ROOK_ON_ADVERSARY_KING_OPEN_FILE_BONUS = 300;

const int BLOCKED_ROOK_PENALTY = -500;

void RookScore(){
	BITBOARD tmp,tmp2;
	int sq,f_mask;

	int rooks_on_seventh;

	/*
		Rook score :
			* big bonus if two rooks on 7th and enemy king on 8th rank depending on number
			of enemy pawns on 7th
			* bonus for rooks connected on a file
			* bonus for rooks on open file
			* bonus for rooks on semi open file
			* rook mobility : bonus for each square that the rooks attacks
	*/

	if (board.piece_counter[WHITE][ROOK]){
		tmp = board.wRooks;

		rooks_on_seventh = 0;

		while (tmp){
			sq = FirstOne(tmp);
			ClearFirstOne(tmp);

			if (Rank(sq) == RANK7)
				rooks_on_seventh++;

			f_mask = 1 << File(sq);

			if ((pawns_on_file[WHITE] & f_mask) == 0){
				if ((pawns_on_file[BLACK] & f_mask) == 0){ // open file
					white_score += ROOK_ON_OPEN_FILE_BONUS;
				} else {	// semiopen file
					white_score += ROOK_ON_SEMIOPEN_FILE_BONUS;
				}
			}
	
			tmp2 = AttacksRook(sq);
			white_score += ROOK_MOBILITY_BONUS * BitCount(tmp2);

			if (board.piece_counter[WHITE][ROOK] > 1)
			if (tmp2 & plus8dir[sq] & board.wRooks)
				white_score += CONNECTED_ROOKS_BONUS;
		}

		if (rooks_on_seventh){
			if (board.black_king_sq >= A8){
				int n_pawns = BitCount(board.bPawns & rank_mask[RANK7]);
				white_score += (rooks_on_seventh == 1) ? rook_on_7th_bonus_onerook[n_pawns] : rook_on_7th_bonus_tworooks[n_pawns];
			} else {
				white_score += (rooks_on_seventh == 1) ? ROOK_ON_7TH_LINE_BONUS2_ONE_ROOK : ROOK_ON_7TH_LINE_BONUS2_TWO_ROOKS;
			}
		}
	}

	if (board.piece_counter[BLACK][ROOK]){
		tmp = board.bRooks;

		rooks_on_seventh = 0;

		while (tmp){
			sq = FirstOne(tmp);
			ClearFirstOne(tmp);

			if (Rank(sq) == RANK2)
				rooks_on_seventh++;

			f_mask = 1 << File(sq);

			if ((pawns_on_file[BLACK] & f_mask) == 0){
				if ((pawns_on_file[WHITE] & f_mask) == 0){ // open file
					black_score += ROOK_ON_OPEN_FILE_BONUS;
				} else {	// semiopen file
					black_score += ROOK_ON_SEMIOPEN_FILE_BONUS;
				}
			}
	
			tmp2 = AttacksRook(sq);
			black_score += ROOK_MOBILITY_BONUS * BitCount(tmp2);

			if (board.piece_counter[BLACK][ROOK] > 1)
			if (tmp2 & minus8dir[sq] & board.bRooks)
				black_score += CONNECTED_ROOKS_BONUS;
		}

		if (rooks_on_seventh){
			if (board.white_king_sq <= H1){
				int n_pawns = BitCount(board.wPawns & rank_mask[RANK2]);
				black_score += (rooks_on_seventh == 1) ? rook_on_7th_bonus_onerook[n_pawns] : rook_on_7th_bonus_tworooks[n_pawns];
			} else {
				black_score += (rooks_on_seventh == 1) ? ROOK_ON_7TH_LINE_BONUS2_ONE_ROOK : ROOK_ON_7TH_LINE_BONUS2_TWO_ROOKS;
			}
		}
	}

	
	// blocked rook
	if (File(board.white_king_sq) >= FILEE){
		if (PieceOnSquare(H1) == ROOK || PieceOnSquare(G1) == ROOK)
			if (board.white_king_sq == G1 || board.white_king_sq == F1)
				white_score += BLOCKED_ROOK_PENALTY;
	} else {
		if (PieceOnSquare(A1) == ROOK || PieceOnSquare(B1) == ROOK)
			if (board.white_king_sq == B1 || board.white_king_sq == C1)
				white_score += BLOCKED_ROOK_PENALTY;
	}


	if (File(board.black_king_sq) >= FILEE){
		if (PieceOnSquare(H8) == -ROOK || PieceOnSquare(G8) == -ROOK)
			if (board.black_king_sq == G8 || board.black_king_sq == F8)
				black_score += BLOCKED_ROOK_PENALTY;
	} else {
		if (PieceOnSquare(A8) == -ROOK || PieceOnSquare(B8) == -ROOK)
			if (board.black_king_sq == B8 || board.black_king_sq == C8)
				black_score += BLOCKED_ROOK_PENALTY;
	}
}


/*
	Knights Score 
		* bonus for proximity to enemy king 
		* bonus if the knight is situated in a square that can't be attacked by
		enemy pawns, this bonus is even bigger if the knight is protected by a pawn
*/


inline int KingDistance(int sq, int king_sq){
	return Max(Abs(File(sq) - File(king_sq)), Abs(Rank(sq) - Rank(king_sq)));
}


extern BITBOARD knight_outpost_mask[2][64];


const int KNIGHT_DISTANCE_TO_KING_BONUS = 10;

const int KNIGHT_OUTPOST_BONUS = 100;
const int KNIGHT_PROTECTED_BY_PAWN_BONUS = 100;
const int KNIGHT_PROTECTED_BY_TWO_PAWNS_BONUS = 200;

/*
	Knight score :
		* bonus for a knight that can't be attacked by an enemy pawn
		* extra bonus if that knight is protected by a pawn
		* more bonus if the knight is protected by 2 pawns, because
		capturing it will make one protected passed pawn
*/

void KnightScore(){
	BITBOARD tmp;
	int sq,n;

	tmp = board.wKnights;
	while(tmp){
		sq = FirstOne(tmp);
		ClearFirstOne(tmp);
		white_score += KNIGHT_DISTANCE_TO_KING_BONUS * (4-KingDistance(sq, board.black_king_sq));
		
		if (sq >= A5) {
			if ((board.bPawns & knight_outpost_mask[WHITE][sq]) == 0){
				white_score += KNIGHT_OUTPOST_BONUS;

				n = BitCount(bpawn_attacks[sq] & board.wPawns);
				if (n == 1)
					white_score += KNIGHT_PROTECTED_BY_PAWN_BONUS;
				else if (n > 1)
					white_score += KNIGHT_PROTECTED_BY_TWO_PAWNS_BONUS;
			}
		}
	}

	tmp = board.bKnights;
	while(tmp){
		sq = FirstOne(tmp);
		ClearFirstOne(tmp);
		black_score += KNIGHT_DISTANCE_TO_KING_BONUS * (4-KingDistance(sq, board.white_king_sq));
		
		if (sq <= H4) {
			if ((board.wPawns & knight_outpost_mask[BLACK][sq]) == 0){
				black_score += KNIGHT_OUTPOST_BONUS;

				n = BitCount(wpawn_attacks[sq] & board.bPawns);
				if (n == 1)
					black_score += KNIGHT_PROTECTED_BY_PAWN_BONUS;
				else if (n > 1)
					black_score += KNIGHT_PROTECTED_BY_TWO_PAWNS_BONUS;
			}
		}
	}

}





char wking_defects[64] = {	1, 0, 2, 3, 3, 2, 0, 1,
							1, 1, 2, 3, 3, 2, 1, 1,
							3, 3, 3, 3, 3, 3, 3, 3,
							4, 4, 4, 4, 4, 4, 4, 4,
							5, 5, 5, 5, 5, 5, 5, 5,
							6, 6, 6, 6, 6, 6, 6, 6,
							7, 7, 7, 7, 7, 7, 7, 7,
							8, 8, 8, 8, 8, 8, 8, 8};
char bking_defects[64] = {	8, 8, 8, 8, 8, 8, 8, 8,
							7, 7, 7, 7, 7, 7, 7, 7,
							6, 6, 6, 6, 6, 6, 6, 6,
							5, 5, 5, 5, 5, 5, 5, 5,
							4, 4, 4, 4, 4, 4, 4, 4,
							3, 3, 3, 3, 3, 3, 3, 3,
							1, 1, 2, 3, 3, 2, 1, 1,
							1, 0, 2, 3, 3, 2, 0, 1};

const int KING_SAFETY_OPEN_FILE = 4;

const int king_safety_lut[64] = {   
		   0,  45, 150, 250, 380, 450, 550, 770, 
		 820, 845, 870, 900, 930, 960, 990,1020, 
		1050,1080,1110,1140,1170,1200,1240,1500,
		1500,1500,1500,1500,1500,1500,1500,1500,
		1500,1500,1500,1500,1500,1500,1500,1500,
		1500,1500,1500,1500,1500,1500,1500,1500,
		1500,1500,1500,1500,1500,1500,1500,1500,
};

const BITBOARD knight_mask_b = OneShl(G5)|OneShl(H5)|OneShl(F5)|OneShl(H6)|OneShl(G6);
const BITBOARD knight_mask_w = OneShl(G4)|OneShl(H4)|OneShl(F4)|OneShl(H3)|OneShl(G3);


extern BITBOARD file_fgh_mask, file_abc_mask;
 
void AddKingSafety(){
	int f;
	
	int safety_w, safety_b;

	int pawns = pawns_on_file[WHITE] | pawns_on_file[BLACK];

	if (board.black_pieces >= 4 || (board.bQueens && board.black_pieces >= 2)){
		safety_w = wking_defects[board.white_king_sq];

		f = File(board.white_king_sq);

		if ((board.castle_flags & (WHITE_SCASTLE | WHITE_LCASTLE)) == 0){
			if (f >= FILEE){
				safety_w += pscore->kside_defects[WHITE];
			
				if (f == FILEE){
					if ((pawns & (1 << FILED)) == 0) safety_w += KING_SAFETY_OPEN_FILE>>1;
					if ((pawns & (1 << FILEE)) == 0) safety_w += KING_SAFETY_OPEN_FILE;
					if ((pawns & (1 << FILEF)) == 0) safety_w += KING_SAFETY_OPEN_FILE>>1;
				} 
			} else if (f <= FILED){
				safety_w += pscore->qside_defects[WHITE];
			
				if (f == FILED){
					if ((pawns & (1 << FILEC)) == 0) safety_w += KING_SAFETY_OPEN_FILE>>1;
					if ((pawns & (1 << FILED)) == 0) safety_w += KING_SAFETY_OPEN_FILE;
					if ((pawns & (1 << FILEE)) == 0) safety_w += KING_SAFETY_OPEN_FILE>>1;
				} 
			}
		} else {
			if ((board.castle_flags & (WHITE_SCASTLE | WHITE_LCASTLE)) != (WHITE_SCASTLE | WHITE_LCASTLE))
				if (board.castle_flags & WHITE_SCASTLE)
					safety_w += pscore->kside_defects[WHITE];
				else if (board.castle_flags & WHITE_LCASTLE)
					safety_w += pscore->qside_defects[WHITE];
		}

		// king on open/halfopen file
		
		if (((1 << f) & pawns_on_file[WHITE]) == 0)	{
			if (board.piece_counter[BLACK][QUEEN] && f == File(FirstOne(board.bQueens)))
				safety_w += 8;
			else
				safety_w += 4;
		}
		/*
		if (f >= FILEF){
			if (board.bKnights & knight_mask_w) safety_w += 2;
			if (board.bQueens & file_fgh_mask) safety_w += 2;
			if (board.bRooks & file_fgh_mask) safety_w += 2;
		} else if (f >= FILEC){
			if (board.bQueens & file_abc_mask) safety_w += 2;
			if (board.bRooks & file_abc_mask) safety_w += 2;
		}
		*/
		if (((1 << f) & pawns_on_file[BLACK]) == 0) safety_w += 2;
		

		if (board.piece_counter[BLACK][QUEEN] == 0) safety_w >>= 1;

		white_score -= king_safety_lut[safety_w];
	}

	if (board.white_pieces >= 4 || (board.wQueens && board.white_pieces >= 2)){
		f = File(board.black_king_sq);

		safety_b = bking_defects[board.black_king_sq];

		if ((board.castle_flags & (BLACK_SCASTLE | BLACK_LCASTLE)) == 0){
			if (f >= FILEE){
				safety_b += pscore->kside_defects[BLACK];

				if (f == FILEE){
					if ((pawns & (1 << FILED)) == 0) safety_b += KING_SAFETY_OPEN_FILE>>1;
					if ((pawns & (1 << FILEE)) == 0) safety_b += KING_SAFETY_OPEN_FILE;
					if ((pawns & (1 << FILEF)) == 0) safety_b += KING_SAFETY_OPEN_FILE>>1;
				} 
			} else if (f <= FILED){
				safety_b += pscore->qside_defects[BLACK];
			
				if (f == FILED){
					if ((pawns & (1 << FILEC)) == 0) safety_b += KING_SAFETY_OPEN_FILE>>1;
					if ((pawns & (1 << FILED)) == 0) safety_b += KING_SAFETY_OPEN_FILE;
					if ((pawns & (1 << FILEE)) == 0) safety_b += KING_SAFETY_OPEN_FILE>>1;
				} 
			}
		} else {
			if ((board.castle_flags & (BLACK_SCASTLE | BLACK_LCASTLE)) != (BLACK_SCASTLE | BLACK_LCASTLE))
				if (board.castle_flags & BLACK_SCASTLE)
					safety_b += pscore->kside_defects[BLACK];
				else if (board.castle_flags & BLACK_LCASTLE)
					safety_b += pscore->qside_defects[BLACK];
		}

		
		if (((1 << f) & pawns_on_file[BLACK]) == 0) {
			if (board.piece_counter[WHITE][QUEEN] && f == File(LastOne(board.wQueens)))
				safety_b += 8;
			else
				safety_b += 4;
		}

		/*		
		if (f >= FILEF){
			if (board.wKnights & knight_mask_b) safety_b += 2;
			if (board.wQueens & file_fgh_mask) safety_b += 2;
			if (board.wRooks & file_fgh_mask) safety_b += 2;
		} else if (f >= FILEC){
			if (board.wQueens & file_abc_mask) safety_b += 2;
			if (board.wRooks & file_abc_mask) safety_b += 2;
		}
		*/
		
		if (((1 << f) & pawns_on_file[WHITE]) == 0) safety_b += 2;
		

		if (board.piece_counter[WHITE][QUEEN] == 0) safety_b >>= 1;

		
		black_score -= king_safety_lut[safety_b];
	}
}










const BITBOARD mask_bishop1 = Or(OneShl(B2),OneShl(D2));
const BITBOARD mask_bishop2 = Or(OneShl(E2),OneShl(G2));
const BITBOARD mask_bishop3 = Or(OneShl(B7),OneShl(D7));
const BITBOARD mask_bishop4 = Or(OneShl(E7),OneShl(G7));


const int passed_pawn_multiplier[8] = {0, 0, 0, 0, 2, 4, 3, 0};

extern int grouped_pawns_bonus[255];

const int wpawn_storm_bonus_lut[8] = {0, 0, 0, 150, 200, 250, 0, 0};
const int bpawn_storm_bonus_lut[8] = {0, 0, 250, 200, 150, 0, 0, 0};

#if 1
const int connected_rank_bonus_ur[3] = {300, 1000, 1500};
const int connected_rank_bonus_r[3] = {300, 1000, 3000};
const int connected_rank_bonus_dr[3] = {300, 1000, 1500};
#else
const int connected_rank_bonus_ur[3] = {0, 000, 00};
const int connected_rank_bonus_r[3] = {0, 000, 000};
const int connected_rank_bonus_dr[3] = {0, 000, 00};
#endif


extern BITBOARD mask_rank567;
extern BITBOARD mask_rank234;

const int PAWN_MAJORITY_BONUS = 100;


int PawnMajorityScore(){
	// pawn majority
	int n_w,n_b;
	int score = 0;
	
	n_w = BitCount(pawns_on_file[WHITE] & 15);
	n_b = BitCount(pawns_on_file[BLACK] & 15);

	if (n_w > n_b)
		score += PAWN_MAJORITY_BONUS;
	else if (n_w < n_b)
		score -= PAWN_MAJORITY_BONUS;

	n_w = BitCount((pawns_on_file[WHITE] >> 4) & 15);
	n_b = BitCount((pawns_on_file[BLACK] >> 4) & 15);

	if (n_w > n_b)
		score += PAWN_MAJORITY_BONUS;
	else if (n_w < n_b)
		score -= PAWN_MAJORITY_BONUS;

	return score;
}



int EvalWhitePawns(){
	int score = 0;
	int sq;
	int i,f,x;
	register BITBOARD tmp;

	for (i = 0; i < 8; i++){
		x = BitCount(board.wPawns & file_mask[i]);
		if (x > 1)
			score -= DOUBLED_PAWN_PENALTY * (x-1);
	}

	//PrintBitboard(board.bKing);
	//cout <<endl;
	//PrintBitboard(free_pawn_ctrl[WHITE][H7]);
	//cout << endl;

	tmp = board.wPawns;
	while (tmp){
		sq = FirstOne(tmp);
		ClearFirstOne(tmp);

		f = File(sq);
		
		if (And(board.wPawns, isolated_pawn_mask[f]) == 0)
			score -= ISOLATED_PAWN_PENALTY;
		else if (And(board.wPawns,backward_pawn_mask[WHITE][sq])==0){
			if ((pawns_on_file[BLACK] & (1 << (File(sq))))==0)
			if (And(wpawn_attacks[sq+8], board.bPawns))
				score -= BACKWARDS_PAWN_PENALTY;
		}

		if (sq >= A5)
		if (And(board.bPawns, free_pawn_ctrl[WHITE][sq]) == 0)
			pscore->passed_pawns |= OneShl(sq);
	}

	score += grouped_pawns_bonus[pawns_on_file[WHITE]];

	return score;
}



int EvalBlackPawns(){
	int score = 0;
	int sq;
	int i,f,x;
	register BITBOARD tmp;

	for (i = 0; i < 8; i++){
		x = BitCount(board.bPawns & file_mask[i]);
		if (x > 1)
			score -= DOUBLED_PAWN_PENALTY * (x-1);
	}
	
	tmp = board.bPawns;
	while (tmp){
		sq = FirstOne(tmp);
		ClearFirstOne(tmp);

		f = File(sq);
		
		if (And(board.bPawns, isolated_pawn_mask[f]) == 0)
			score -= ISOLATED_PAWN_PENALTY;
		else if (And(board.bPawns,backward_pawn_mask[BLACK][sq])==0){
			if ((pawns_on_file[WHITE] & (1 << (File(sq))))==0)
			if (And(bpawn_attacks[sq-8], board.wPawns))
				score -= BACKWARDS_PAWN_PENALTY;
		}

		if (sq < A5)
		if (And(board.wPawns, free_pawn_ctrl[BLACK][sq]) == 0)
			pscore->passed_pawns |= OneShl(sq);
	}

	score += grouped_pawns_bonus[pawns_on_file[BLACK]];

	
	return score;
}


extern BITBOARD king_pawn_mask1[2][64];	// mask for pawns on rank = (king's rank + 2)
extern BITBOARD king_pawn_mask2[2][64];	// mask for pawns on rank > (king's rank + 2)


const int QUEENSIDE = 0;
const int CENTER = 1;
const int KINGSIDE = 2;


const int BitCount3[8] = {0, 1, 1, 2, 1, 2, 2, 3};
const int NegBitCount3[8] = {3, 2, 2, 1, 2, 1, 1, 0};

const BITBOARD mask_F3G3 = OneShl(F3)|OneShl(G3);
const BITBOARD mask_A3B3 = OneShl(A3)|OneShl(B3);
const BITBOARD mask_F6G6 = OneShl(F6)|OneShl(G6);
const BITBOARD mask_A6B6 = OneShl(A6)|OneShl(B6);

const int missing_pawns_multiplier[8] = { 0, 1, 2, 3, 4, 5, 6, 7};
const int openfiles_multiplier[4] = { 0, 8, 13, 16};
const int halfopenfiles_multiplier[4] = { 0, 3, 7, 9};


void ComputeKingScore(){
	int file;

	int kmiss_w, kmiss_b, qmiss_w, qmiss_b;
	int kside_open_files;
	int kside_halfopen_files_w, kside_halfopen_files_b;
	int qside_open_files;
	int qside_halfopen_files_w, qside_halfopen_files_b;

	kmiss_w = kmiss_b = qmiss_w = qmiss_b = 0;
	kside_halfopen_files_w = kside_halfopen_files_b = 0;
	qside_halfopen_files_w = qside_halfopen_files_b = 0;

	kside_open_files = NegBitCount3[(pawns_on_file[WHITE] | pawns_on_file[BLACK]) >> 5];
	qside_open_files = NegBitCount3[(pawns_on_file[WHITE] | pawns_on_file[BLACK])  & 7];
	kside_halfopen_files_w = BitCount3[(((~pawns_on_file[WHITE]) & pawns_on_file[BLACK]) >> 5) & 7];
	kside_halfopen_files_b = BitCount3[(((~pawns_on_file[BLACK]) & pawns_on_file[WHITE]) >> 5) & 7];
	qside_halfopen_files_w = BitCount3[((~pawns_on_file[WHITE]) & pawns_on_file[BLACK]) & 7];
	qside_halfopen_files_b = BitCount3[((~pawns_on_file[BLACK]) & pawns_on_file[WHITE]) & 7];

	for (file = 0; file < 3; file++){
		if (pawns_on_file[WHITE] & (1 << (7-file))){
			if (Not(board.wPawns & OneShl(H2-file))){
				kmiss_w++;
				if (Not(board.wPawns & OneShl(H3-file))) kmiss_w++;
				if (file == 1) kmiss_w++;
			}
		}

		if (pawns_on_file[BLACK] & (1 << (7-file))){
			if (Not(board.bPawns & OneShl(H7-file))){
				kmiss_b++;
				if (Not(board.bPawns & OneShl(H6-file))) kmiss_b++;
				if (file == 1) kmiss_b++;
			}
		} 

		if (pawns_on_file[WHITE] & (1 << file)){
			if (Not(board.wPawns & OneShl(A2+file))){
				qmiss_w++;
				if (Not(board.wPawns & OneShl(A3+file))) qmiss_w++;
				if (file == 1) qmiss_w++;
			}
		}

		if (pawns_on_file[BLACK] & (1 << file)){
			if (Not(board.bPawns & OneShl(A7+file))){
				qmiss_b++;
				if (Not(board.bPawns & OneShl(A6+file))) qmiss_b++;
				if (file == 1) qmiss_b++;
			}
		} 
	}

	// special case : two pawns advanced to F3G3 or A3G3 / F6G6 A6B6 are not as dangerous as the seem
	if (kmiss_w) if ((board.wPawns & mask_F3G3) == mask_F3G3) kmiss_w--;
	if (qmiss_w) if ((board.wPawns & mask_A3B3) == mask_A3B3) qmiss_w--;
	if (kmiss_b) if ((board.bPawns & mask_F6G6) == mask_F6G6) kmiss_b--;
	if (qmiss_b) if ((board.bPawns & mask_A6B6) == mask_A6B6) qmiss_b--;

	// defect count
	pscore->kside_defects[WHITE] =	missing_pawns_multiplier[kmiss_w] + 
									openfiles_multiplier[kside_open_files] + 
									halfopenfiles_multiplier[kside_halfopen_files_w]+
									(halfopenfiles_multiplier[kside_halfopen_files_b]>>1);
	pscore->qside_defects[WHITE] =	missing_pawns_multiplier[qmiss_w] + 
									openfiles_multiplier[qside_open_files] + 
									halfopenfiles_multiplier[qside_halfopen_files_w]+
									(halfopenfiles_multiplier[qside_halfopen_files_b]>>1);
	pscore->kside_defects[BLACK] =	missing_pawns_multiplier[kmiss_b] + 
									openfiles_multiplier[kside_open_files] + 
									halfopenfiles_multiplier[kside_halfopen_files_b]+
									(halfopenfiles_multiplier[kside_halfopen_files_w]>>1);
	pscore->qside_defects[BLACK] =	missing_pawns_multiplier[qmiss_b] + 
									openfiles_multiplier[qside_open_files] + 
									halfopenfiles_multiplier[qside_halfopen_files_b]+
									(halfopenfiles_multiplier[qside_halfopen_files_w]>>1);
}



extern BITBOARD pawn_kingsquare[2][64];

const int PAWN_RACE_BONUS[8] = {0, 0, 0, 0, 1500, 2500, 3000, 0};
const int PAWN_RACE_BONUS2[8] = {0, 0, 0, 0, 1000, 1500, 2000, 0};


void PawnRaceScore(int wtm){
	register BITBOARD tmp;
	int sq;


	// add passed pawn score

	tmp = pscore->passed_pawns;
	while (tmp){
		sq = FirstOne(tmp);
		ClearFirstOne(tmp);

		if (sq >= A5) {	// white pawn
			if (board.black_king_sq != sq+8)
				if (And(board.bKing, free_pawn_ctrl[WHITE][sq]) == 0)
					white_score += PASSED_PAWN_BONUS * passed_pawn_multiplier[Rank(sq)];
		} else {		// black pawn
			if (board.white_king_sq != sq-8)
				if (And(board.wKing, free_pawn_ctrl[BLACK][sq]) == 0)
					black_score += PASSED_PAWN_BONUS * passed_pawn_multiplier[7-Rank(sq)];
		}
	}

	// score for white if no pieces left for the black
	if (board.black_material <= 1){
		tmp = board.wPawns;
		while (tmp){
			sq = FirstOne(tmp);
			ClearFirstOne(tmp);
			if (And(board.bPawns,free_pawn_ctrl[WHITE][sq]) == 0){
				if (And(pawn_kingsquare[WHITE][wtm ? sq : sq+8], board.bKing) == 0)
					if (board.black_material == 0){
						white_score += PAWN_RACE_BONUS[Rank(sq)];
					} else {
						white_score += PAWN_RACE_BONUS2[Rank(sq)];
					}
			}
		}
	}

	// score for black if no pieces left for the white
	if (board.white_material <= 1){
		tmp = board.bPawns;
		while (tmp){
			sq = FirstOne(tmp);
			ClearFirstOne(tmp);
			if (And(board.wPawns,free_pawn_ctrl[BLACK][sq]) == 0){
				if (And(pawn_kingsquare[BLACK][(!wtm) ? sq : sq-8], board.wKing) == 0)
					if (board.white_material == 0){
						black_score += PAWN_RACE_BONUS[7-Rank(sq)];
					}else {
						black_score += PAWN_RACE_BONUS2[7-Rank(sq)];
					}
			}
		}
	}
}



const int BOTH_BISHOPS_BONUS = 150;


const int bishop_bonus_lut[17] = { 1000, 1000, 1000, 900, 800, 700, 600, 400, 350, 325, 300, 275, 275, 250, 250, 150, 150};

const int rook_exchange_lut[17] = { 0, -300, 0, 100, 150, 200, 900, 1100, 1200, 1300, 1400, 1450, 1500, 1500, 1500, 1500, 1500};

int BishopScore(){
	int score = 0;

	if (board.piece_counter[WHITE][BISHOP] > 1)
		score += bishop_bonus_lut[board.white_pawns + board.black_pawns];
	if (board.piece_counter[BLACK][BISHOP] > 1)
		score -= bishop_bonus_lut[board.white_pawns + board.black_pawns];

	return score;
}

int RookForLightFiguresTradeScore(){
	int score = 0;

	int wminors = board.piece_counter[WHITE][KNIGHT] + board.piece_counter[WHITE][BISHOP];
	int bminors = board.piece_counter[BLACK][KNIGHT] + board.piece_counter[BLACK][BISHOP];

	if (wminors != bminors){
		int wmajors = board.piece_counter[WHITE][ROOK] + 2*board.piece_counter[WHITE][QUEEN];
		int bmajors = board.piece_counter[BLACK][ROOK] + 2*board.piece_counter[BLACK][QUEEN];
		
		if (Abs(wmajors-bmajors) == 1){
			if (wminors == bminors+2)
				score += rook_exchange_lut[board.white_pawns + board.black_pawns];	
			else if (bminors == wminors+2)
				score -= rook_exchange_lut[board.white_pawns + board.black_pawns];
		}
	}


	return score;
}


const int KING_IN_FRONT_OF_PAWN_PENALTY = -300;

void PenalizeKingInFrontOfThePawn(){
	if (board.white_pieces <= 2 && board.black_pieces <= 2){
		if (board.white_king_sq >= A6)
			if (PieceOnSquare(board.white_king_sq - 8) == PAWN)
				white_score += KING_IN_FRONT_OF_PAWN_PENALTY;

		if (board.black_king_sq <= H3)
			if (PieceOnSquare(board.black_king_sq + 8) == -PAWN)
				black_score += KING_IN_FRONT_OF_PAWN_PENALTY;
	}
}




extern int castle_penalty[2];


void CastlePenalty(){
	const BITBOARD wcastle_mask =	OneShl(D1)|OneShl(E1)|OneShl(F1)|
									OneShl(D2)|OneShl(E2)|OneShl(F2)|OneShl(C2)|
									OneShl(D3)|OneShl(E3)|OneShl(F3)|OneShl(C3);
	const BITBOARD bcastle_mask =	OneShl(D8)|OneShl(E8)|OneShl(F8)|
									OneShl(D7)|OneShl(E7)|OneShl(F7)|OneShl(C7)|
									OneShl(D6)|OneShl(E6)|OneShl(F6)|OneShl(C6);

	if (castle_penalty[WHITE]){
		if (!WhiteCastled())
			if ((board.castle_flags & (WHITE_SCASTLE|WHITE_LCASTLE)) == 0)
				white_score += board.bQueens ? 1*castle_penalty[WHITE] : castle_penalty[WHITE];
	}

	if (castle_penalty[BLACK]){
		if (!BlackCastled())
			if ((board.castle_flags & (BLACK_SCASTLE|BLACK_LCASTLE)) == 0)
				black_score += board.wQueens ? 1*castle_penalty[BLACK] : castle_penalty[BLACK];
	}
}


extern int undeveloped_pieces_penalty[2];

void UndevelopedPiecesPenalty(){
	int n;

	const int undeveloped_table[5] = {0, -10, -400, -750, -1000};

	if (undeveloped_pieces_penalty[WHITE]){
		n = 0;
		if (PieceOnSquare(B1) == KNIGHT) n++;
		if (PieceOnSquare(C1) == BISHOP) n++;
		if (PieceOnSquare(G1) == KNIGHT) n++;
		if (PieceOnSquare(F1) == BISHOP) n++;
		white_score += undeveloped_table[n];
	};
	if (undeveloped_pieces_penalty[BLACK]){
		n = 0;
		if (PieceOnSquare(B8) == -KNIGHT) n++;
		if (PieceOnSquare(C8) == -BISHOP) n++;
		if (PieceOnSquare(G8) == -KNIGHT) n++;
		if (PieceOnSquare(F8) == -BISHOP) n++;
		black_score += undeveloped_table[n];
	};
}


const int EXCHANGE_BONUS0 = 2000;

void ExchangePieces(){
	if (board.white_material == 0 && board.black_material == 0){
		if (board.white_pawns > board.black_pawns && board.material >= PAWN_V)
			white_score += EXCHANGE_BONUS0;
		else if (board.black_pawns > board.white_pawns && board.material <= -PAWN_V)
			black_score += EXCHANGE_BONUS0;
	}
}




#define LAZY_EVAL(margin) {																		\
	register int score = (wtm ? (white_score - black_score) : (black_score - white_score));	\
	if (score + margin < alpha || score - margin > beta) return score;			\
}


int MAX_POSN_SCORE = 2000; // reseted to 2000 each time the board is initialized


int trade_bonus[17] = {
	1250,1140,1020, 950, 820, 700, 600, 500,	/*   0 -  7 */
	 350, 220, 140,  70,  55,  35,  20,  10,	/*   8 - 15 */
	   0										/*  16 - 16 */
};

void MaterialAdvantage(){

	if (board.material >= 1000)
		white_score += trade_bonus[board.white_material+board.black_material] - trade_bonus[board.white_pawns+board.black_pawns];
	else if (board.material <= -1000)
		black_score += trade_bonus[board.white_material+board.black_material] - trade_bonus[board.white_pawns+board.black_pawns];

	/*
	int MD, PA, MT;

	if (board.material == 0) return; // no need for this

	MD = board.material;
	PA = MD > 0 ? board.white_pawns : board.black_pawns;
	MT = +board.piece_counter[WHITE][KNIGHT]*3
		 +board.piece_counter[WHITE][BISHOP]*3
		 +board.piece_counter[WHITE][ROOK]*5
		 +board.piece_counter[WHITE][QUEEN]*9
		 +board.piece_counter[BLACK][KNIGHT]*3
		 +board.piece_counter[BLACK][BISHOP]*3
		 +board.piece_counter[BLACK][ROOK]*5
		 +board.piece_counter[BLACK][QUEEN]*9;
	MT *= 1000;
		
		
	white_score += Min(24000, MD) + MD * PA * (80000-MT)/(64000 * (PA+1));	
	*/


}


int Board::Evaluate(int alpha, int beta, int ply, int wtm){

	white_score = 0;
	black_score = 0;

	white_score = board.material + board.piecetable_score;
	black_score = 0;


	pscore = ProbePawnHT();	
	if (!pscore){
		pscore = &pawn_score;

		pscore->passed_pawns = 0;

		ComputePawnMasks();
		pscore->pawn_score  = EvalWhitePawns() - EvalBlackPawns();
		pscore->pawn_score += PawnMajorityScore();
		ComputeKingScore();

		pscore->pawns_on_files = pawns_on_file[WHITE] | (pawns_on_file[BLACK] << 8);

		AddPawnHT(pawn_score);
	} else{							
		pawns_on_file[WHITE] = pscore->pawns_on_files & 255;
		pawns_on_file[BLACK] = pscore->pawns_on_files >> 8;
	}

	white_score += pscore->pawn_score;

	AddKingSafety();

	//return (wtm ? (white_score - black_score) : (black_score - white_score));


	//MaterialAdvantage();

	LAZY_EVAL(MAX_POSN_SCORE);
	int tscore = white_score-black_score;

	PawnRaceScore(wtm);

	ExchangePieces();

	UndevelopedPiecesPenalty();

	white_score += RookForLightFiguresTradeScore();

	white_score += BishopScore();

	PenalizeKingInFrontOfThePawn();

	KnightScore();
	
	RookScore();

	CastlePenalty();

	if (Abs(white_score-black_score - tscore) > MAX_POSN_SCORE)
		MAX_POSN_SCORE = Abs(white_score-black_score - tscore);

	stats.evals++;

	return (wtm ? (white_score - black_score) : (black_score - white_score));
}

