#include "bitboard.h"
#include "init.h"
#include "board.h"
#include <iostream>
#include "pgn.h"
#include "book.h"
#include "input.h"
#include "prophet.h"
#include <string.h>
#include <stdlib.h>
#include "movelist.h"
#include "hash.h"
#include "config.h"

using namespace std;

void InitCommandLine();


/*
	Command line options : 

		Help screen :
			-help
		
		Analysing :

			-analyse <epd filename>
			-outfile <filename to write results to>
			-maxtime <max time for a position to be analysed in seconds (float)>
			-maxply <max ply to analyse a position>
			-tag <engine configutration description when analysing EPDs>
			-epdmode : epd output

		Playing:
		
			-xboard 

		General options

			-hashsize <x : integer between 5 and 25> hash size will be 2^x
			-tbdir xxx : path to the tablebases files : default tbs/

		Building an opening book

			it's done in two steps : pgn -> temp file -> opening book

			-pgnsouce <name of the file to build the book from>
			-tempfile <name of the temporary file default temp.dat>
			-max_book_moves <number of moves to include from each match into the book, default 10>
			-buildbook <name of the temp file>

*/

#include "swap.h"

#define DEFAULT_HASH_SIZE 18

#include "contra.h"

static struct {
	char* epd_name;
	char* outfile;
	int xboard;
	int analyse;
	int maxtime;
	int maxply;
	char* description;
	int hash_size;
	char* pgn_name;
	char* temp_name;
	int build_book;
	int max_book_moves;
	char* tbdir;
} options = { NULL, NULL, 1, 0, 26000, 8, "unknown configuration", DEFAULT_HASH_SIZE, NULL, "temp.dat", 0, 12, "E:/tbs"};


extern void PrintTables(ostream& stream);

#define WIN32_LEAN_AND_MEAN
#include <windows.h>


#include "search.h"
#include "move.h"

#include "movelist.h"


int PawnScore(int& pawns_on_files);

void PrintPawnsOnFiles(int pof){
	int i;

	for (i = 0; i < 8; i++)
		cout <<  ((pof & (1 << i)) ? "1" : "0");
	cout << endl;
	for (i = 0; i < 8; i++)
		cout <<  ((pof & (1 << (i+8))) ? "1" : "0");
	cout << endl;
}

void dummy(void* dummy){
	Sleep(5000);
	input_flags.time_left = -1;
}

#include <windows.h>
#include <process.h>

#include "tablebases.h"

void LearnBook(char* fname, char* compiled_file);

extern void DumpBook(int wtm, int ply = 0);




/*
	#if 0
	
	for (int wm = 0; wm <= 8; wm+=1)
	for (int bm = 0; bm <= 8; bm+=1)
	for (int wp = 0; wp <= 8; wp++)
	for (int bp = 0; bp <= 8; bp++){
		//if (Abs(wm-bm) > 5) continue;

		if (wm != bm || wp != bp) continue;

		int score = 0;
		//if (wm+wp > bm+bp)
			score += trade_bonus[wm+bm]-trade_bonus[wp+bp];
		//else if (bm+bp > wp+wm)
		//	score -= trade_bonus[wm+bm]-trade_bonus[wp+bp];

		cout << "score = " << score << "	";
		cout << "wm = " << wm << "	bm = " << bm << "	wp = " << wp << "	bp = " << bp;
		cout << endl;
	}

	return 0;
	
	#endif
*/



extern void test_bitboards();


extern int trade_bonus[];

int __cdecl main(int argc, char* argv[]){

    setbuf(stdout, NULL);
    setbuf(stdin, NULL);
    setvbuf(stdout, NULL, _IONBF, 0);
    setvbuf(stdin, NULL, _IONBF, 0);   
	
    DeleteFile("evaluate.debug");


	input_flags.limit_depth = 0;
	input_flags.limit_time = 0;
	input_flags.use_books = 1;


#if 0

	InitHT(DEFAULT_HASH_SIZE);
	InitTables();	
	InitTB(options.tbdir);

	#if 0
	ReadBook("book.bin", "index.dat");


	LearnBook("learn.txt", "learn.bin");
	
	board.Reset();
	ResetMoveList(WHITE);

	MakeMove(D2, D4, 0);
	MakeMove(G8, F6, 0);
	MakeMove(C2, C4, 0);
	MakeMove(E7, E6, 0);

	board.Print();

	ProbeBook(WHITE);
	#endif

	#if 0
	//test_bitboards();
	
	{ char str[256]; cin >> str; return 0; }
	#endif
	

	input_flags.time_left = 265000; // 0.5 minute for each position


	//board.Init("4q2k/2b5/3p1p2/R3P2r/3P1P2/6B1/8/3KQ3 w - - 0 1");
	//board.Init("r3nrk1/2p2p1p/p1p1b1p1/2NpPq2/3R4/P1N1Q3/1PP2PPP/4R1K1 w - - bm g4; id \"WAC_orig.0023\";");
	//board.Init("r1bq1rk1/1pp2pbp/3p2p1/p1nPp3/2P1N3/4BN2/PPQ1BPPP/R4RK1 b - - 0 11");
	//board.Init("r4rk1/1pp2pbp/3p2p1/p1nPp3/2P1bP1N/2N1B3/PP2B1PP/R4RK1 b - f3 0 14");
	//board.Init("r4rk1/1pp2pbp/3p2p1/p1nbp3/2P2P1N/2N1B3/PP2B1PP/R4RK1 w - - 0 15");
	//board.Init("r4rk1/1pp2pbp/3p2p1/p1nPpb2/2P4N/2N1B3/PP2BPPP/R4RK1 b - - 0 13");
	//board.Init("7r/7P/1pk5/p7/6R1/6K1/8/8 b");
	//board.Init("2kr1b1r/4npp1/pq6/3p3p/1pPN1Bb1/1P6/1P2NPPP/R1Q1R1K1 b - - 0 20");
	//board.Init("7K/8/2p5/3p4/3P4/4P3/8/7k b - - 0 1");
	//board.Init("r1bq1rk1/pp1p1pp1/2n1pn1p/8/Qb2PB2/2N2N2/PPP2PPP/2KR1B1R w - - 0 10");
	//board.Init("r1bq1rk1/pp1p1pp1/2n1pn1p/1B6/Q3PB2/2P2N2/P1P2PPP/2KR3R b - - 0 11");
	
	//board.Init("rnbq1rk1/pppn1ppp/8/8/Q1BNP3/8/PP1N1PPP/R3K2R w KQ - 0 10");
	//board.Init("rnbq1rk1/pppn1ppp/8/8/Q1BNP3/8/PP1N1PPP/2KR3R b - - 0 10");
	//board.Init("rnbq1rk1/pppn1ppp/8/8/Q1BNP3/8/PP1N1PPP/R4RK1 b - - 0 10");
	//board.Init("rnbq1rk1/pppn1ppp/8/8/Q1BNP3/8/PP1N1PPP/1R2K2R b K - 0 1");

	//board.Init("r1bqk2r/pp2bppp/3p1n2/2p1p3/4P3/1PNP1N1P/1PP2PP1/R1BQ1RK1 b kq - 0 9");

	//board.Init("1rb1k2r/3p1ppp/pp1q1n2/4p3/P3P3/2PBBP2/2PQ2PP/1R3RK1 b k - 0 16");
	//board.Init("1rbk3r/3p1ppp/pp1q1n2/4p3/P3P3/2PBBP2/2PQ2PP/1R3RK1 b - - 0 1");

	//board.Init("4k3/3p4/1PR2p2/4p1pr/4P2p/1P3P2/r5PP/3R2K1 b - - 0 32");
	//board.Init("r1b1k2r/3p1p1p/p1n1p3/1p4p1/4P3/1PP5/P2NBPPP/R4RK1 b kq - 0 16");
	//board.Init("r1b2rk1/3p1p1p/p1n1p3/1p4p1/4P3/1PP5/P2NBPPP/R4RK1 w - - 0 17");

	//board.Init("rnb1kb1r/pp4pp/1qp2p2/7Q/3PN3/8/PPP3PP/R3KBNR b KQkq - 0 9");
	//board.Init("rnb1kb1r/pp5p/1qp2pp1/7Q/3PN3/8/PPP3PP/R3KBNR b KQkq - 0 1");

	//board.Init("3r2br/p2R2p1/k7/1p3n2/2P1p2p/K5P1/2P2P2/R7 w - - 0 30");
	
	//board.Init("8/k7/3p4/p2P1p2/P2P1P2/8/8/K7 w - - bm Kb1");

	
	//board.Init("rq2r1k1/5pp1/p7/4bNP1/1p2P2P/5Q2/PP4K1/5R1R w - - bm Nxg7;");

	//board.Init("8/p7/P5p1/7p/7P/4kpK1/8/8 w");

	//board.Init("8/8/3p4/7p/P2p4/3K2P1/7b/3k4 b - - 0 2");

	//board.Init("3B4/8/2B5/1K6/8/8/3p4/3k4 w - - bm Ka6");

	//board.Init("6k1/2b2p1p/ppP3p1/4p3/PP1B4/5PP1/7P/7K w - - bm Bxb6;");

	//board.Init("1nbq1r1k/3rbp1p/p1p1pp1Q/1p6/P1pPN3/5NP1/1P2PPBP/R4RK1 w - - bm Nfg5;");

	//board.Init("K7/8/8/8/8/8/1k6/R7 w - - 0 8");

	//board.Init("2b1k2r/1p1p1p1p/r1nPp1pQ/p3q3/3p2P1/2P5/PP1P1PBP/R1B2RK1 b k - 0 16");


	//board.Init("r3rbk1/ppq2ppp/2b1pB2/8/6Q1/1P1B3P/P1P2PP1/R2R2K1 w - - bm Bxh7;");


	//board.Init("r3kbnr/ppp2ppp/3p4/3Nn3/2B1Pp1q/5Q2/PPPP2PP/R1B2K1R w kq - 0 9");
	//board.Init("8/k1b5/P4p2/1Pp2p1p/K1P2P1P/8/3B4/8 w - - bm b6+;");

	//board.Init("1r4nk/1p1qb2p/3p1r2/p1pPp3/2P1Pp2/5P1P/PP1QNBRK/5R2 b; bm Qxh3!!");

	//board.Reset();


	//board.Evaluate(1);
	//return 0;

	/* 1 */
	board.Init("r1b2rk1/ppbn1ppp/4p3/1QP4q/3P4/N4N2/5PPP/R1B2RK1 w - - bm c6");
	
	/* 2 */
	//board.Init("r3nrk1/2p2p1p/p1p1b1p1/2NpPq2/3R4/P1N1Q3/1PP2PPP/4R1K1 w - - bm g4");
	
	/* 3 */
	//board.Init("r2q2k1/pp1rbppp/4pn2/2P5/1P3B2/6P1/P3QPBP/1R3RK1 w - - bm c6");

	/* 4 */
	//board.Init("2kr3r/pp1q1ppp/5n2/1Nb5/2Pp1B2/7Q/P4PPP/1R3RK1 w - - bm Nxa7+");

	
	
	//board.CheckMirroredEvalEPD("c:/chess/freyr/ecm98.epd", "c:/chess/freyr/ecm98.mirror");
	//return 0;
	 

	//board.Init("3rr1k1/pb2qpp1/n6p/2p5/4p3/P1N1P3/1PQN1PPP/2R2RK1 w - - 0 18");
	//board.Init("5rk1/Rb3pp1/7p/Q1p5/4p3/P1N1P3/1P1q1PPP/6K1 w - - 0 31");

	//board.Init("2kr3r/pp1n1pb1/3p1p2/2pP3p/6Pq/1BB2P1P/PPQ2P2/R1K4R w  h6");

	//board.Init("r2r2k1/pb2bp1p/1q2p3/4P1p1/n7/PN3N1P/1PP1QPP1/1R1R2K1 b - - 0 20");
	//board.Init("5r2/1b2k2p/3bp3/5PP1/Bn1Q3P/1N6/1Pr5/1K3R2 b - - bm Rxb2+;");

	//board.Init("3r2k1/ppr1p2p/4P1p1/1B6/5P2/1R2B2P/P2b2P1/7K w - - 0 31");

	
	//board.Init("6k1/p5p1/1pb2P2/3p2R1/P1pP3P/2r2B2/2P3K1/8 b - - 0 38");


	#if 0
		board.Print();

		cout << board.Evaluate(-INF, INF, 0, board.side_to_move) << endl;;
		board.Mirror();
		cout << board.Evaluate(-INF, INF, 0, board.side_to_move) << endl;;

		return 0;
	#endif

	//board.Print();
	//cout << board.Evaluate(-INF, INF, 0, WHITE) << endl;;


	#if 0


	#if 1

	ReadBook("book.bin", "index.dat");
	board.Reset();

	#if 0
	HASHCODE parent = board.hashcode;

	int move = D2 + (D4 << 6) + (PAWN << 12);

	board.Make(move, 0, WHITE);

	HASHCODE child = board.hashcode;

	HASHCODE xx = Shl(Shr(parent,48),48) + Shr(Shl(child,16),16);

	Print(parent); cout << endl;
	Print(child); cout << endl;
	Print(xx); cout << endl;

	board.Reset();
	#endif
	
	board.Init("rnbqkbnr/pp1pp1pp/8/2pP1p2/2P5/8/PP2PPPP/RNBQKBNR b KQkq c3 0 3");

	//ProbeBook(BLACK);
	DumpBook(BLACK, 0);

	return 0;
	
	#else	
	
	/*
	InitHT(17);
	InitTables();
	*/
	board.Reset();
	
	/*
	ReadBook("book.bin", "index.dat");

	int dummy;

	input_flags.time_left = 260000;
	
	board.Analyse(WHITE, dummy);
	*/
	
	//ParsePgn("clarendn.pgn", "temp.dat", 5);
	BuildBook(options.temp_name, "book.bin", "index.dat");
	return 0;

	#endif

	{char str[1024]; cin >> str; return 0;}

	#endif

	/*
	board.Init("8/2P2k1p/p4pp1/2p1p3/P6P/5PP1/3b4/7K w");
	board.Print();
	PrintMove(7858); 
	cout << board.IsValidMove(7858,1) << endl;
	
	{char str[1024];
	cin >> str;}
	return 0;
	*/
	
	input_flags.time_left = 260000;
	//input_flags.time_left = 500;
	
	int next_move;
	board.Analyse(board.side_to_move, next_move, 37);


	//PrintCounterMoves();

	//{char str[1024]; cin >> str;}


	return 0;
#endif	

	input_flags.epd_mode = 0;
	
	for (int x = 0; x <argc; x++)
		cout << argv[x] << endl;

	for (int i = 1; i < argc; i++){
		if (argv[i][0] == '-'){
			if (strstr(argv[i], "-analyse")){
				if (i+1 < argc){
					options.analyse = 1;
					options.epd_name = argv[++i];
					options.xboard = 0;					
				}
			} else if (strstr(argv[i], "-xboard")){
				options.analyse = 0;
				options.xboard = 1;					
			} else if (strstr(argv[i], "-maxtime")){
				if (i+1 < argc)
					options.maxtime = (int)(atof(argv[++i]) * 100.0f);
			} else if (strstr(argv[i], "-maxply")){
				if (i+1 < argc)
					options.maxply = atoi(argv[++i]);
			} else if (strstr(argv[i], "-outfile")){					
				if (i+1 < argc)
					options.outfile = argv[++i];;
			} else if (strstr(argv[i], "-tag")){
				if (i+1 < argc)
					options.description = argv[++i];
			} else if (strstr(argv[i], "-hashsize")){
				if (i+1 < argc){
					options.hash_size = atoi(argv[++i]);
				}
				cout << "hashsize : " << options.hash_size << endl;
			} else if (strstr(argv[i], "-pgnsource")){
				if (i+1 < argc){
					options.pgn_name = argv[++i];
				}
			} else if (strstr(argv[i], "-tempfile")){
				if (i+1 < argc){
					options.temp_name = argv[++i];
				}
			} else if (strstr(argv[i], "-buildbook")){
				options.build_book = 1;
			} else if (strstr(argv[i], "-max_book_moves")){
				if (i+1 < argc){
					options.max_book_moves = atoi(argv[++i]);
				}
			} else if (strstr(argv[i], "-epdmode")){
				input_flags.epd_mode = 1;
			
			} else if (strstr(argv[i], "-tbdir")){
				if (i+1 < argc){
					options.tbdir = argv[++i];
				}
			} else {
				cout << "Unknown parameter : " << argv[i] << endl;
			}
		}
	}

	/*
	ParsePgn("bigbook.pgn", "temp.dat");

	//BuildBook("d:/chess/0/temp.dat", "d:/chess/0/book.bin", "d:/chess/0/index.dat");
	//BuildBook("temp.dat", "book.bin", "index.dat");
	*/
	
	if (options.build_book){
		InitHT(17);
		InitTables();
		BuildBook(options.temp_name, "book.bin", "index.dat");
		return 0;
	} else if (options.pgn_name){
		InitHT(17);
		InitTables();
		ParsePgn(options.pgn_name, options.temp_name, options.max_book_moves);
		return 0;
	}
	

	InitHT(options.hash_size);
	InitTables();
	InitTB(options.tbdir);

	if (options.analyse && options.epd_name){
		board.AnalyseEPD(options.epd_name, options.outfile, options.maxtime, options.maxply, options.description);	
	} else if (options.xboard){
		//InitCommandLine();
	
		board.Reset();
		ReadBook("book.bin", "index.dat");

		InitCommandLine();
	}

#if 0
	#if 0
		BuildBook("temp.dat", "book.bin", "index.dat");
	#else	
		board.Reset();

		ReadBook("book.bin", "index.dat");
		ProbeBook(1);
	#endif
#endif


	return 0;
}





unsigned int SqFindKing(int* sq){
	//cout << "SqFindKing" << endl;
	
	//for (int x = 0; x < 64; x++) cout << sq[x] << " ";
	//cout << endl;

	for (int i = 0; i < 64; i++, sq++){
		if (*sq == 6){
			//cout << i << endl;
			return i;
		}
	}

	cout << "Not found !" << endl;
	return 127;
}

unsigned int SqFindFirst(int* sq, int p){
	//cout << "SqFindFirst " << p << endl;

	//for (int x = 0; x < 64; x++) cout << sq[x] << " ";
	//cout << endl;

	for (int i = 0; i < 64; i++, sq++)
		if (*sq == p){
			//cout << i << endl;
			return i;
		}
	cout << "Not found !" << endl;
	return 127;
}
