#include "board.h"
#include "move.h"
#include "prophet.h"
#include <iostream>
#include <string.h>

using namespace std;

//#define VERIFY_MAKEMOVE

//#define VERIFY_HASH

//#define VERIFY_MATERIAL

//#define VERIFY_PIECETABLE

//#define VERIFY_PIECECOUNTERS

extern int piecetable[2][7][64];

extern const int piece_v[7];

extern void PrintTables(ostream& stream);

void Board::Make(int move, int ply, int wtm){
	int from = From(move);
	int to = To(move);
	int piece = Piece(move);
	int capture = Captured(move);
	int promote = Promote(move);

	if (capture) 
		piece_counter[1-wtm][capture]--;
	if (promote){
		piece_counter[wtm][promote]++;
		piece_counter[wtm][PAWN]--;
	}

	#ifdef VERIFY_PIECETABLE
		Assert(board.piecetable_score == EvaluateStaticPosition());
	#endif

	save_struct[ply].ep_move = (capture && sq[to] == 0);

	/*
	board.Print();
	
	
	cout << "  Make move ";
	PrintMove(move);
	cout << "(" << from << " " << to << " " << piece << " " << promote << " " << board.castle_flags <<  ")" << endl;
	/*/
	/*
	for (int i = 0; i < ply; i++) PrintMove(current_move[i]);
	PrintMove(move);
	cout << endl;
	//board.Print();
	//::Print(board.pawn_hashcode); cout << endl;
	*/

	//Assert(board.pawn_hashcode == board.ComputePawnHashcode(wtm));

	UpdateHash(move, wtm);

	/*
	cout << "  Make move ";
	PrintMove(move);
	cout << "(" << from << " " << to << " " << piece << " " << promote << " " << board.castle_flags <<  ")" << endl;
	*/

	SetClearBoardBit(from, to);

	ep_square = 0;

	if (wtm){
		SetClear(white_pieces, from, to);

		// clearing the captured piece must be done before moving the capturing piece
		if (sq[to] != FREE)
			Clear(black_pieces, to);

		if (capture > PAWN)
			black_material--; 
		else if (capture == PAWN)
			black_pawns--;

		material += piece_v[capture];

		if (promote){
			white_pawns--;
			white_material++;
			material += piece_v[promote] - piece_v[PAWN];
		}

		if (capture != 0){
			if (sq[to] == FREE)  // en passant
				board.piecetable_score += piecetable[BLACK][capture][to-8];
			else 
				board.piecetable_score += piecetable[BLACK][capture][to];
		}

		board.piecetable_score += piecetable[WHITE][piece][to] - piecetable[WHITE][piece][from];

		if (piece == KING) {	// castle ?
			if (from == E1){
				if (to == G1)
					board.piecetable_score += piecetable[WHITE][ROOK][F1] - piecetable[WHITE][ROOK][H1];
				else if (to == C1)
					board.piecetable_score += piecetable[WHITE][ROOK][D1] - piecetable[WHITE][ROOK][A1];
			}
		}
		if (promote)
			board.piecetable_score += piecetable[WHITE][promote][to] - piecetable[WHITE][piece][to];

		switch(capture){
		case KNIGHT:
			Clear(bKnights, to);
			break;
		case BISHOP:
			Clear(bBishops, to);
			Clear(bishops_queens, to);
			break;
		case ROOK:
			if (to == H8) castle_flags &= ~BLACK_SCASTLE;
			else if (to == A8) castle_flags &= ~BLACK_LCASTLE;
			Clear(bRooks, to);
			Clear(rooks_queens, to);
			break;
		case QUEEN:
			Clear(bQueens, to);
			Clear(rooks_queens, to);
			Clear(bishops_queens, to);
			break;
		case KING:
			cout << "Captured king !" << endl;
			{
			for (int i = 0; i <= ply; i++)
				PrintMove(current_move[i]);
			cout << endl;
			Assert(false);
			}
			break;
		case PAWN:
			if (sq[to] == FREE){ // ep check
				Assert(piece == PAWN && sq[to-8] == -PAWN);
				sq[to-8] = FREE;
				Clear(bPawns, to-8);
				ClearBoardBit(to-8);
				Clear(black_pieces, to-8);
			} else {
				Clear(bPawns, to);
			}
		}

		sq[to] = piece;
		sq[from] = FREE;

		switch(piece){
		case KNIGHT:
			SetClear(wKnights, from, to);
			break;

		case BISHOP:			
			SetClear(wBishops, from, to);
			SetClear(bishops_queens, from, to);
			break;

		case ROOK:
			// clear castle flags if necessary
			if (from == H1)
				castle_flags &= ~WHITE_SCASTLE;
			else if (from == A1)
				castle_flags &= ~WHITE_LCASTLE;

			SetClear(wRooks, from, to);
			SetClear(rooks_queens, from, to);
			break;

		case QUEEN:
			SetClear(wQueens, from, to);
			SetClear(rooks_queens, from, to);
			SetClear(bishops_queens, from, to);
			break;

		case KING :
			SetClear(wKing, from, to);
			white_king_sq = to;	

			// castle
			if (from == E1){
				castle_flags &= ~(WHITE_SCASTLE | WHITE_LCASTLE) ;
				if (to == G1) {
					Assert(sq[H1] == ROOK);
					SetClearBoardBit(H1, F1);
					SetClear(wRooks, H1, F1);
					SetClear(rooks_queens, H1, F1);
					SetClear(white_pieces, H1, F1);
					sq[H1] = 0; 
					sq[F1] = ROOK;
					board.castle_flags |= WHITE_HAS_CASTLED;
				} else if (to == C1){
					Assert(sq[A1] == ROOK);
					SetClearBoardBit(A1, D1);
					SetClear(wRooks, A1, D1);
					SetClear(rooks_queens, A1, D1);
					SetClear(white_pieces, A1, D1);
					sq[A1] = 0; 
					sq[D1] = ROOK;
					board.castle_flags |= WHITE_HAS_CASTLED;
				}
			}	
			break;

		case PAWN :
			SetClear(wPawns, from, to);		
			if (to == from+16) // set ep flag if necessary
				ep_square = from+8;
			else 
			if (to >= 56){ // pawn promotion
				sq[to] = promote;
				Clear(wPawns, to);
				switch(promote){
				case QUEEN:
					Set(wQueens, to);
					Set(bishops_queens, to);
					Set(rooks_queens, to);
					break;
				case KNIGHT:
					Set(wKnights, to);
					break;
				case ROOK:
					Set(wRooks, to);
					Set(rooks_queens, to);
					break;
				case BISHOP:
					Set(wBishops, to);
					Set(bishops_queens, to);
					break;
				default:
					//board.Init(epd_string);
					//board.Print();
					//cout << epd_string << endl;
					PrintMove(move);
					cout << this->IsValidMove(move, 1);
					cout << endl;
					Assert(false);
				}
			}
			break;

		default:
			Assert(false);
		}

	} else {

		SetClear(black_pieces, from, to);

		// clearing the captured piece must be done before moving the capturing piece
		if (sq[to] != FREE)
			Clear(white_pieces, to);

		if (capture > PAWN)
			white_material--; 
		else if (capture == PAWN)
			white_pawns--;

		material -= piece_v[capture];

		if (promote){
			black_pawns--;
			black_material++;
			material -= piece_v[promote] - piece_v[PAWN];
		}


		if (capture != 0){
			if (sq[to] == FREE)  // en passant
				board.piecetable_score -= piecetable[WHITE][capture][to+8];
			else 
				board.piecetable_score -= piecetable[WHITE][capture][to];
		}
		board.piecetable_score -= (piecetable[BLACK][piece][to] - piecetable[BLACK][piece][from]);
		if (piece == KING) {	// castle ?
			if (from == E8){
				if (to == G8)
					board.piecetable_score -= piecetable[BLACK][ROOK][F8] - piecetable[BLACK][ROOK][H8];
				else if (to == C8)
					board.piecetable_score -= piecetable[BLACK][ROOK][D8] - piecetable[BLACK][ROOK][A8];
			}
		}

		if (promote)
			board.piecetable_score -= piecetable[BLACK][promote][to] - piecetable[BLACK][piece][to];


		switch(capture){
		case KNIGHT:
			Clear(wKnights, to);
			break;
		case BISHOP:
			Clear(wBishops, to);
			Clear(bishops_queens, to);
			break;
		case ROOK:
			if (to == H1) castle_flags &= ~WHITE_SCASTLE;
			else if (to == A1) castle_flags &= ~WHITE_LCASTLE;
			Clear(wRooks, to);
			Clear(rooks_queens, to);
			break;
		case QUEEN:
			Clear(wQueens, to);
			Clear(rooks_queens, to);
			Clear(bishops_queens, to);
			break;
		case KING:
			cout << "Captured king !" << endl;
			break;
		case PAWN:
			if (sq[to] == FREE){ // ep check
				Assert(piece == PAWN && sq[to+8] == PAWN);
				sq[to+8] = FREE;
				Clear(wPawns, to+8);
				ClearBoardBit(to+8);
				Clear(white_pieces, to+8);
			} else {
				Clear(wPawns, to);
			}
		}

		sq[to] = -piece;
		sq[from] = FREE;

		switch(piece){
		case KNIGHT:
			SetClear(bKnights, from, to);
			break;

		case BISHOP:			
			SetClear(bBishops, from, to);
			SetClear(bishops_queens, from, to);
			break;

		case ROOK:
			// clear castle flags if necessary
			if (from == H8)
				castle_flags &= ~BLACK_SCASTLE;
			else if (from == A8)
				castle_flags &= ~BLACK_LCASTLE;

			SetClear(bRooks, from, to);
			SetClear(rooks_queens, from, to);
			break;

		case QUEEN:
			SetClear(bQueens, from, to);
			SetClear(rooks_queens, from, to);
			SetClear(bishops_queens, from, to);
			break;

		case KING :
			SetClear(bKing, from, to);
			black_king_sq = to;	

			// castle
			if (from == E8){
				castle_flags &= ~(BLACK_SCASTLE | BLACK_LCASTLE) ;
				if (to == G8) {
					Assert(sq[H8] == -ROOK);
					SetClearBoardBit(H8, F8);
					SetClear(bRooks, H8, F8);
					SetClear(rooks_queens, H8, F8);
					SetClear(black_pieces, H8, F8);
					sq[H8] = 0; 
					sq[F8] = -ROOK;
					board.castle_flags |= BLACK_HAS_CASTLED;
				} else if (to == C8){
					Assert(sq[A8] == -ROOK);
					SetClearBoardBit(A8, D8);
					SetClear(bRooks, A8, D8);
					SetClear(rooks_queens, A8, D8);
					SetClear(black_pieces, A8, D8);
					sq[A8] = 0; 
					sq[D8] = -ROOK;
					board.castle_flags |= BLACK_HAS_CASTLED;
				}
			}	
			break;

		case PAWN :
			SetClear(bPawns, from, to);		
			if (to == from-16)	// set ep flag if necessary
				ep_square = from-8;
			else 
			if (to <= 7) { // pawn promotion
				sq[to] = -promote;
				Clear(bPawns, to);
				switch(promote){
				case QUEEN:
					Set(bQueens, to);
					Set(bishops_queens, to);
					Set(rooks_queens, to);
					break;
				case KNIGHT:
					Set(bKnights, to);
					break;
				case ROOK:
					Set(bRooks, to);
					Set(rooks_queens, to);
					break;
				case BISHOP:
					Set(bBishops, to);
					Set(bishops_queens, to);
					break;
				default:
					Assert(false);
				}
			}
			break;

		default:
			Assert(false);
		}

	}


	#ifdef VERIFY_MAKEMOVE
		Verify("MakeMove");
	#endif

	#ifdef VERIFY_HASH
		board.VerifyHash(1-wtm);
		board.VerifyPawnHash(1-wtm);
	#endif

	#ifdef VERIFY_MATERIAL
		VerifyMaterial();
	#endif

	#ifdef VERIFY_PIECETABLE
		if (board.piecetable_score != EvaluateStaticPosition()){
			
			PrintTables(cout);

			board.Print();
			for (int i = 0; i < ply; i++) PrintMove(current_move[i]);
			PrintMove(move);
			for (i = 0; i < ply; i++) cout << current_move[i] << " ";
			cout << move;
			cout << endl << endl;

			cout << "Current score : " << board.piecetable_score << endl;
			cout << "Should be	   : " << EvaluateStaticPosition() << endl;

			Assert(board.piecetable_score == EvaluateStaticPosition());
		}
	#endif

	#ifdef VERIFY_PIECECOUNTERS
		board.VerifyPieceCounter();
	#endif
}
