#include "search.h"
#include "board.h"
#include "move.h"
#include <iostream>
#include "hash.h"
#include "search.h"
#include "input.h"
#include "swap.h"
#include "killer.h"
#include "hashlist.h"
#include "history.h"
#include "utils.h"
#include <string.h>
#include "tablebases.h"
#include "stats.h"
#include "tree.h"
#include "contra.h"
#include "rootmove.h"

using namespace std;


#define MAX(a,b) (((a) > (b)) ? (a) : (b))


#define MAX_MOVES 128000
int move_list[MAX_MOVES];
int last_move[MAX_PLY];	// pointer in move_list to the last move for each ply

int score[MAX_MOVES];


int current_move[MAX_PLY];
int current_ply = 0;

int pv[MAX_PLY][MAX_PLY];  // principal variation
int pvlen[MAX_PLY];		   // length of each pv


extern HASHCODE whiteblack_hash;		// shortcut for white_hash ^ black_hash

#define TRY_HASHMOVE						1
#define TRY_KILLER1							1
#define TRY_KILLER2							1
#define CLEAR_KILLERS						1
#define ORDER_CAPTURES						1
#define TRY_NULLMOVE						1
#define TRY_PV_MOVE							1
#define TRY_HISTORY							1
#define TRY_IID								1
#define TRY_COUNTERMOVE						1

#define LAST_HISTORY_PLY					4

#define MAX_HISTORY_MOVES_TRIED				4

#define NULLMOVE_DEPTH  (3 * PLY_DEPTH)


// try ETC if ply <= ETC_MAX_PLY and depth >= ETC_MIN_DEPTH
#define ETC_MAX_PLY 66
#define ETC_MIN_DEPTH (3 * PLY_DEPTH)

#define CHECK_EXTENSION						1
#define RECAPTURE_ON_SAME_SQUARE_EXTENSION	0 
#define PAWN_PROMOTED_TO_QUEEN_EXTENSION	0  
#define PAWN_MOVED_TO_7TH_RANK_EXTENSION	0  
#define PAWN_MOVED_TO_6TH_RANK_EXTENSION	0  
#define NULL_THREAT_EXTENSION				1

#define CHECK_EXTENSION_DEPTH				((PLY_DEPTH * 3)/4)
#define RECAPTURE_ON_SAME_SQUARE_DEPTH		((PLY_DEPTH * 1)/2)
#define PAWN_PROMOTED_TO_QUEEN_DEPTH		((PLY_DEPTH * 1)/4)
#define PAWN_PROMOTED_TO_7TH_RANK_DEPTH		((PLY_DEPTH * 1)/4)
#define PAWN_PROMOTED_TO_6TH_RANK_DEPTH		((PLY_DEPTH * 1)/4)
#define NULL_THREAT_DEPTH				      PLY_DEPTH


#define MaterialGain(M) (piece_v[Captured(M)])


inline void SortMoves(int* moves, int* scores, int n){
	int tmp, done;
	int i;

	register int* p_score;
	register int* p_move_list;

	do {
		//for (int j = 0; j < n; j++) cout << scores[j] << " "; cout << endl;

		p_score = scores;
		p_move_list = moves;

		done = 1;
		for (i = 0; i+1 < n; i++, p_score++, p_move_list++){
			if (*p_score < *(p_score+1)){
				tmp = *p_score; *p_score = *(p_score+1); *(p_score+1) = tmp;
				tmp = *p_move_list; *p_move_list = *(p_move_list+1); *(p_move_list+1) = tmp;
				done = 0;
			}
		}
	} while (!done);
}


int MULTIPLIER = 1000;


inline void ScoreCapturesSEE(int* first, int* last, int* score, int wtm){
	static int piece, capture, promote;
	int* score0 = score;

	for (int* m = first; m < last; m++, score++){
		
		piece = piece_v[Piece(*m)];
		capture = piece_v[Captured(*m)];
		promote = piece_v[Promote(*m)];
	
		*score = Swap(From(*m), To(*m), wtm);
		if (promote) 
			*score += promote - piece;

		*score *= MULTIPLIER;
	}

	AddHistoryScores(first, last, score0, wtm);
}

inline void ScoreCapturesNoSEE(int* first, int* last, int* score, int wtm){
	int* score0 = score;
	for (int* m = first; m < last; m++, score++){
		*score++ = (piece_v[Captured(*m)]-piece_v[Piece(*m)]) * MULTIPLIER;
	}
	AddHistoryScores(first, last, score0, wtm);
}

inline void ScoreNonCapturesSEE(int* first, int* last, int* score, int wtm){
	int* score0 = score;
	for (int* m = first; m < last; m++, score++){
		if (m) 
			*score = Swap(From(*m), To(*m), wtm) * MULTIPLIER;
	}
	AddHistoryScores(first, last, score0, wtm);
}

inline void ScoreNonCapturesNoSEE(int* first, int* last, int* score, int wtm){
	//memset(score, 0, (last-first)*sizeof(int));
	FillHistoryScores(first, last, score, wtm);
}


extern int* pv_move;
extern int pv_moves_left;


const int MAX_POSN_SCORE = 3 * PAWN_V;
const int FUTIL_MARGIN = 6 * PAWN_V;
const int RAZOR_MARGIN = 10 * PAWN_V;

#define fab(M) {																		\
	current_ply = ply;																	\
	HashListAdd();																		\
	board.Make(M, ply, wtm);															\
																						\
	int legal_move = !Check(wtm);														\
																						\
	int cutoff = 0;																		\
																						\
	if (legal_move){																	\
		moves_searched++;																\
																						\
		current_move[ply] = M;															\
																						\
		int new_depth = depth- PLY_DEPTH;												\
		if (RECAPTURE_ON_SAME_SQUARE_EXTENSION){										\
			if (ply > 0){																\
 				if (To(M) == To(current_move[ply-1]) && Captured(current_move[ply-1]))	\
					new_depth += RECAPTURE_ON_SAME_SQUARE_DEPTH;						\
			}																			\
		}																				\
																						\
		if (Piece(M) == PAWN){															\
			if (PAWN_PROMOTED_TO_QUEEN_EXTENSION){										\
				if (Promote(M) == QUEEN)												\
					new_depth += PAWN_PROMOTED_TO_QUEEN_DEPTH;							\
			}																			\
			if (wtm){																	\
				if (PAWN_MOVED_TO_7TH_RANK_EXTENSION){									\
					if (Rank(To(M)) == RANK7)												\
						new_depth += PAWN_PROMOTED_TO_7TH_RANK_DEPTH;					\
				} 																		\
				if (PAWN_MOVED_TO_6TH_RANK_EXTENSION){									\
					if (Rank(To(M)) == RANK6)												\
						new_depth += PAWN_PROMOTED_TO_6TH_RANK_DEPTH;					\
				} 																		\
			} else {																	\
				if (PAWN_MOVED_TO_7TH_RANK_EXTENSION){									\
					if (Rank(To(M)) == RANK2)												\
						new_depth += PAWN_PROMOTED_TO_7TH_RANK_DEPTH;					\
				} 																		\
				if (PAWN_MOVED_TO_6TH_RANK_EXTENSION){									\
					if (Rank(To(M)) == RANK3)												\
						new_depth += PAWN_PROMOTED_TO_6TH_RANK_DEPTH;					\
				} 																		\
			}																			\
		}																				\
		/*																				\
		if (1 && new_depth <= 3 * PLY_DEPTH && new_depth > 2*PLY_DEPTH ){												\
			int eval = (wtm ? board.material : -board.material) + MaterialGain(M) + RAZOR_MARGIN; \
			if (eval <= gamma)																		   \
				new_depth -= PLY_DEPTH;																			   \
		}																							   \
		if (1 && new_depth <= 1 * PLY_DEPTH){												\
			int eval = (wtm ? board.material : -board.material) + MaterialGain(M) + MAX_POSN_SCORE; \
			if (eval <= gamma && !Check(1-wtm))																		   \
				cutoff = 1;																			   \
		}																							   \
		if (1 && new_depth <= 2 * PLY_DEPTH && new_depth > PLY_DEPTH ){												\
			int eval = (wtm ? board.material : -board.material) + MaterialGain(M) + FUTIL_MARGIN; \
			if (eval <= gamma && !Check(1-wtm))																		   \
				cutoff = 1;																			   \
		}																							   \
		*/																				\
		/*score = -Search(-beta, -MAX(gamma,alpha), new_depth, ply+1, 1-wtm, 1);		\
		/*/																				\
																						\
		if (!cutoff){																	\
			if (moves_searched == 1){														\
				score = -Search(-beta, -MAX(gamma,alpha), new_depth, ply+1, 1-wtm, 1);		\
			} else {																		\
				score = -Search(-gamma-1, -gamma, new_depth, ply+1, 1-wtm, 1);				\
				if (score > gamma && score < beta){											\
					score = -Search(-beta, -MAX(gamma, alpha), new_depth, ply+1, 1-wtm, 1);	\
				}																			\
			}																				\
		}																			\
	}																					\
																						\
	current_ply = ply;																	\
	board.Unmake(M, ply, wtm);															\
	HashListRemoveLast();																\
																						\
	if (input_flags.time_left <= 0) return 0;											\
	if (legal_move && !cutoff){																	\
		if (score > gamma){																\
			if (!tree[ply+1].same_position_or_rule50)									\
				tree[ply].same_position_or_rule50 = 1;									\
			else																		\
				tree[ply].same_position_or_rule50 = 0;									\
			best_move = M;																\
			gamma = score;																\
			UpdatePV(ply, M);															\
			if (gamma >= beta){															\
				/* don't add to hashtable if rule50 or 3x same position on ply+1 */		\
				if (!tree[ply+1].same_position_or_rule50){								\
					AddHT(gamma, FAIL_HIGH, depth, M);									\
				} 																		\
																						\
				if (ply>0) 																\
					AddCounterMove(current_move[ply-1], M, wtm, depth);						\
																						\
				stats.hashmove_cutoffs++;												\
				if (store_killers) AddKiller(M, ply);									\
				if (store_history) AddHistoryMove(wtm, depth, M);						\
				if (ply == 0) UpdatePV(ply, M);											\
				return beta;															\
			}																			\
		}																				\
	}																					\
}



#define fab_nocheck(M) {																		\
	current_ply = ply;																	\
	HashListAdd();																		\
	board.Make(M, ply, wtm);															\
																						\
	int legal_move = 1;																	\
																						\
	int cutoff = 0;																		\
																						\
	if (legal_move){																	\
		moves_searched++;																\
																						\
		current_move[ply] = M;															\
																						\
		int new_depth = depth- PLY_DEPTH;												\
		/*																				\
		if (1 && new_depth <= 3 * PLY_DEPTH && new_depth > 2*PLY_DEPTH ){												\
			int eval = (wtm ? board.material : -board.material) + MaterialGain(M) + RAZOR_MARGIN; \
			if (eval <= gamma)																		   \
				new_depth -= PLY_DEPTH;																			   \
		}																							   \
		if (1 && new_depth <= 1 * PLY_DEPTH){												\
			int eval = (wtm ? board.material : -board.material) + MaterialGain(M) + MAX_POSN_SCORE; \
			if (eval <= gamma && !Check(1-wtm))																		   \
				cutoff = 1;																			   \
		}																							   \
		if (1 && new_depth <= 2 * PLY_DEPTH && new_depth > PLY_DEPTH ){												\
			int eval = (wtm ? board.material : -board.material) + MaterialGain(M) + FUTIL_MARGIN; \
			if (eval <= gamma && !Check(1-wtm))																		   \
				cutoff = 1;																			   \
		}																							   \
		*/																				\
		if (!cutoff){																	\
			if (moves_searched == 1){														\
				score = -Search(-beta, -MAX(gamma,alpha), new_depth, ply+1, 1-wtm, 1);		\
			} else {																		\
				score = -Search(-gamma-1, -gamma, new_depth, ply+1, 1-wtm, 1);				\
				if (score > gamma && score < beta){											\
					score = -Search(-beta, -MAX(gamma, alpha), new_depth, ply+1, 1-wtm, 1);	\
				}																			\
			}																				\
		}																			\
	}																					\
																						\
	current_ply = ply;																	\
	board.Unmake(M, ply, wtm);															\
	HashListRemoveLast();																\
																						\
	if (input_flags.time_left <= 0) return 0;											\
	if (legal_move && !cutoff){																	\
		if (score > gamma){																\
			if (!tree[ply+1].same_position_or_rule50)									\
				tree[ply].same_position_or_rule50 = 1;									\
			else																		\
				tree[ply].same_position_or_rule50 = 0;									\
			best_move = M;																\
			gamma = score;																\
			UpdatePV(ply, M);															\
			if (gamma >= beta){															\
				/* don't add to hashtable if rule50 or 3x same position on ply+1 */		\
				if (!tree[ply+1].same_position_or_rule50){								\
					AddHT(gamma, FAIL_HIGH, depth, M);									\
				} 																		\
																						\
				if (ply>0) 																\
					AddCounterMove(current_move[ply-1], M, wtm, depth);						\
																						\
				stats.hashmove_cutoffs++;												\
				if (store_killers) AddKiller(M, ply);									\
				if (store_history) AddHistoryMove(wtm, depth, M);						\
				if (ply == 0) UpdatePV(ply, M);											\
				return beta;															\
			}																			\
		}																				\
	}																					\
}




void PrintScore(int* score, int n, char* msg){
	return; // !!!!
	
	cout << "PrintScore " << msg << endl;
	for (int i = 0; i < n; i++)
		cout << "	[" << i << "] " << score[i] << endl;
	cout << "--------------" << endl;
}


void FindFirstMove(int *first,int *last, int*score){
	int* p_score = score+1;
	int* m = first+1;
	int tmp;
	for ( ;m < last; m++, p_score++){
		if (*p_score > *score){
			tmp = *p_score; *p_score = *score; *score = tmp;
			tmp = *m; *m = *first; *first = tmp;
		}
	}
}


int Search(int alpha, int beta, int depth, int ply, int wtm, int do_null){
	int moves_searched;
	int best_move;
	int *last, *m_list, *m;
	int score;
	int hash_move;
	int *p_score;
	int store_history;
	int store_killers;
	int killer1, killer2;
	int pv_move;
	
	// draw by repetition check
	tree[ply].same_position_or_rule50 = 0;
	if (HashListRepetition2x()){
		tree[ply].same_position_or_rule50 = 1;
		return DrawScore(wtm);
	}



	//int gamma = alpha;
	int gamma = -INF;

	pvlen[ply] = 0;

	// return draw if not enought material is left on the board
	if (board.white_material + board.black_material + board.white_pawns + board.black_pawns == 0)
		return 0;

	int in_check = Check(wtm);

	// if in check extend the search by 3/4 ply
	#if CHECK_EXTENSION
		if (in_check){	
			depth += CHECK_EXTENSION_DEPTH;
		}
	#endif


	if (depth < PLY_DEPTH){
		score = QSearch(alpha, beta, ply, depth, wtm);
		return score;
	}


	if (ply > 0 && BitCount(board.all) <= 5 /*&& (ply == 0 || (ply > 0 && Captured(current_move[ply-1])))*/)
	if (EGTBProbe(ply, wtm, &score)){
		//board.Print();
		//cout << score << endl << endl;
		//return score;
		if (score <= -MIN_MATE_SCORE) 
			return score+ply*10;
		else if (score >= MIN_MATE_SCORE)
			return score-ply*10;
		else
			return score;
	}

	stats.nodes++;

	SaveBoardInfo(ply);

	CheckSearchTime();

	store_history = 0;
	store_killers = 0;
	moves_searched = 0;
	best_move = 0;
	killer1 = 0;
	killer2 = 0;
	pv_move = 0;

	m_list = (ply > 0) ? FirstMove(ply) : move_list;
	last_move[ply] = m_list-move_list;



	#if TRY_PV_MOVE
		if (pv_moves_left > 0){
			pv_move = *(::pv_move);
			::pv_move++;
			pv_moves_left--;

			if (pv_move && board.IsValidMove(pv_move, wtm))
				fab(pv_move);
		}
	#endif



	hash_move = 0;
	if (ply > 0)
	if (ProbeHT(alpha, beta, depth, score, hash_move))	{
		if (score == -MATE) score = -MATE+ply*10;
		return score;
	}


	// Null move
	// It isn't tried if side to move has no other pieces than king and pawns or ply==0
	#if TRY_NULLMOVE
		if (do_null && !in_check)
		if (beta != -INF && alpha+1 == beta && ply > 0)
		if ((wtm ? board.white_material : board.black_material) > 0){
		
			int null_depth = (depth <= 6*PLY_DEPTH) ? 3*PLY_DEPTH : 4*PLY_DEPTH;

			board.hashcode ^= whiteblack_hash;

			int ep_square = board.ep_square;
			board.ep_square = 0;

			score = -Search(-beta, -beta+1, depth - null_depth, ply+1, 1-wtm, 0);

			board.hashcode ^= whiteblack_hash;
			board.ep_square = ep_square;

			if (input_flags.time_left <= 0) return 0;

			if (score >= beta){
				AddHT(beta, FAIL_HIGH, depth, 0);
				return beta;
			}

			#if NULL_THREAT_EXTENSION
				if (score <= -MIN_MATE_SCORE){
					depth += NULL_THREAT_DEPTH;
				}
			#endif
		}
	#endif


	last_move[ply] = m_list-move_list;
	if (hash_move && hash_move != pv_move && board.IsValidMove(hash_move, wtm)) 
		fab(hash_move);

	/*
		Internal Iterative Deepening
		
		we are in a part of the tree where alpha != beta-1
		and we have no hash_move to try
		so we try to find a good move by doing a shallower search
	*/

	#if TRY_IID

	if (hash_move == 0 && do_null && depth > 2*PLY_DEPTH && pv_move == 0 && alpha+1 != beta ){
		score = Search(alpha, beta, depth - 2*PLY_DEPTH, ply+1, wtm, 1);

		if (score <= alpha){
			score = Search(-INF, beta, depth - 2*PLY_DEPTH, ply+1, wtm, 1);
		}

		if (input_flags.time_left <= 0) return 0;

		if (pvlen[ply+1] > 0){
			pv_move = pv[ply+1][0];
			if (!board.IsValidMove(pv_move, wtm)){
				#if 0
				cout << "Non valid pv move !" << endl;
				board.Print();
				PrintMove(pv_move); cout << endl;
				#endif
				pv_move = 0;
			} else {
				fab(pv_move);
			}
		}
	}

	#endif


	// Try counter-move
	int counter_move = 0;


	#if TRY_COUNTERMOVE	
		counter_move = (ply > 0) ? GetCounterMove(current_move[ply-1], wtm) : 0;

		if (counter_move == hash_move || counter_move == pv_move)
			counter_move = 0;

		if (counter_move && board.IsValidMove(counter_move, wtm)){
			fab(counter_move);
		}
	#endif


	last = board.GenerateCaptures(m_list, wtm);
	last_move[ply] = (last-1)-move_list;





	//TEST !!
	#if 1
	if (in_check){
		last = board.GenerateCheckEvasions(m_list, wtm);

		if (last == m_list)
			return -MATE+10*ply;

		// one reply check ? extend depth
		#if 1
		if (last - m_list == 1)
			depth += 1*PLY_DEPTH - CHECK_EXTENSION_DEPTH;
		#endif
		


		last_move[ply] = last-move_list;


		/* Score moves 
				Capturing the checking piece (i.e. capture moves) go first
				Interposing pieces go second
				And moving the king goes third
		*/

		killer1 = GetKiller1(ply);
		killer2 = GetKiller2(ply);
		
		int *score_start = ::score + (m_list- move_list);
		int *p_score;

		FillHistoryScores(m_list, last, score_start, wtm);
		
		for (m = m_list, p_score = score_start; m < last; m++,p_score++){
			
			if (Captured(*m))
				*p_score = Swap(From(*m), To(*m), wtm) * 1000000000;
			else {
				;//*p_score = 0;
				if (*m == killer1) *p_score += 5000000;
				if (*m == killer2) *p_score += 2500000;
			} 
		}

		//SortMoves(m_list, score_start, last-m_list);

		for (m = m_list, p_score = score_start; m < last; m++, p_score++){
			FindFirstMove(m,last, p_score);
			if (*m == hash_move) continue;
			if (*m == pv_move) continue;
			if (*m == counter_move) continue;


			if (Captured(*m)){
				store_killers = 0;
				store_history = 0;
			} else {
				store_killers = 1;
				store_history = 1;
			}

			fab_nocheck(*m);
		}

		if (gamma > alpha){
			AddHT(gamma, EXACT, depth, best_move);
		} else 
			AddHT(gamma, FAIL_LOW, depth, best_move);

		if (ply > 0)
			AddCounterMove(current_move[ply-1], 0, wtm, depth);


		return gamma;
	}
	#endif


	ScoreCapturesSEE(m_list, last, ::score + (m_list - move_list), wtm);	

	//PrintScore(::score + (m_list - move_list), last-m_list, "after ScoreCapturesSEE");

	#if ORDER_CAPTURES
		SortMoves(m_list, ::score + (m_list - move_list), last-m_list);
	#endif

	//PrintScore(::score + (m_list - move_list), last-m_list, "after SortMoves");


	// gainful captures	
	for (m = m_list, p_score = ::score + (m_list - move_list); m < last; m++, p_score++){
		if (*p_score <= 0) continue;
		if (*m == hash_move) continue;
		if (*m == counter_move) continue;
		if (*m == pv_move) continue;
		if (*m == killer1) continue;
		if (*m == killer2) continue;
		if (*m == 0) continue;

		fab(*m);
		*m = 0;
	}


	// even captures
	#if 0
	for (m = m_list, p_score = ::score + (m_list - move_list); m < last; m++, p_score++){
		if (*p_score != 0) continue;
		if (*m == 0) continue;
		if (*m == counter_move) continue;
		if (*m == hash_move) continue;
		if (*m == pv_move) continue;
		if (*m == killer1) continue;
		if (*m == killer2) continue;
		fab(*m);
		*m = 0;
	}
	#endif


	#if CLEAR_KILLERS
		ClearKillerScore(ply+1);
	#endif

	// killer1
	#if TRY_KILLER1
		killer1 = GetKiller1(ply);
		if (killer1 == hash_move) killer1 = 0;
		if (killer1 == pv_move) killer1 = 0;
		if (killer1 == counter_move) killer1 = 0;
		if (killer1 && board.IsValidMove(killer1, wtm)) 
			fab(killer1);
	#endif

	// killer2
	#if TRY_KILLER2
		killer2 = GetKiller2(ply);
		if (killer2 == hash_move || killer2 == killer1) killer2 = 0;
		if (killer2 == pv_move) killer2 = 0;
		if (killer2 == counter_move) killer2 = 0;
		if (killer2 && board.IsValidMove(killer2, wtm)) 
			fab(killer2);
	#endif

	store_killers = 1;
	store_history = 1;


	int* first_noncapture = last;
	last = board.GenerateNonCaptures(last, wtm);
	last_move[ply] = (last-1)-move_list;




	ScoreNonCapturesNoSEE(first_noncapture, last, ::score + (first_noncapture - move_list), wtm);	
	

	int noncaptures_tried = 0;

	for (m = m_list; m < last; m++){
		FindFirstMove(m, last, ::score+(m-move_list));

		if (*m == 0) continue;
		if (*m == hash_move) continue;
		if (*m == killer1) continue;
		if (*m == counter_move) continue;
		if (*m == killer2) continue;
		if (*m == pv_move) continue;

		fab(*m);
	}


	if (moves_searched == 0){
		int mate = in_check;

		if (mate) {
			//!!!AddHT(-MATE, EXACT, 50 * PLY_DEPTH, 0);
			return -MATE+10*ply;
		}
		else return 0;
	}

	if (gamma > alpha){
		AddHT(gamma, EXACT, depth, best_move);

		AddKiller(best_move, ply);
		AddHistoryMove(wtm, depth, best_move);
	} else 
		AddHT(gamma, FAIL_LOW, depth, best_move);

	if (ply > 0)
		AddCounterMove(current_move[ply-1], 0, wtm, depth);
	
	return gamma;
}


//#undef UpdatePV





int SearchRoot(int alpha, int beta, int depth, int wtm){
	int moves_searched = 0;
	int best_move;
	int i;
	int score;
	int nodes0, nodes;

    
	root_info.dont_check_time = 0;
	
	int ply = depth/PLY_DEPTH;
	
	#if 0
	cout << "Move list is : ";
	for (i = 0; i < root_info.n_moves; i++)
		PrintMove(root_info.m[i].move);
	cout<<endl;
	#endif


	root_info.singular = 0;

	root_info.SortMovesByNodes();

	//alpha = root_alpha;
	//beta = root_beta;

	int alpha0 = alpha;

	pvlen[0] = 0;

	SaveBoardInfo(0);

	// return draw if not enought material is left on the board
	if (board.white_material + board.black_material + board.white_pawns + board.black_pawns == 0)
		return 0;

	stats.nodes++;

	CheckSearchTime();

	
//	int nodes0, nodes;

	moves_searched = 0;
	best_move = 0;

	last_move[0] = 0;

	if (depth > 2*PLY_DEPTH){
		alpha = root_info.m[0].score - 250;
		beta = root_info.m[0].score + 250;
	}	

	int is_pv = 1;
	
	for (i = 0; i < root_info.n_moves; i++){
		int m = root_info.m[i].move;
		

		/*
		if (i == 0){
			::pv_move = root_info.m[0].pv+1;
			::pv_moves_left = root_info.m[0].pvlen-1; //root_pv_length[m-move_list];
		}
		*/
		/*		
			::pv_move = root_info.m[i].pv+1;
			::pv_moves_left = root_info.m[i].pvlen-1; //root_pv_length[m-move_list];
		*/
		
		current_ply = 0;
		HashListAdd();
		board.Make(m, 0, wtm);	
																						
		moves_searched++;
																						
		current_move[0] = m;
																						
																						
		int new_depth = depth-PLY_DEPTH;	

		nodes0 = stats.nodes + stats.qnodes;

		score = -Search(-beta, -alpha, new_depth, 1, 1-wtm, 1);
		
	

		if (score <= alpha){ 
			if (is_pv){ // fail low on first move
				beta = score;
				alpha = -INF;
				score = -Search(-beta, -alpha, new_depth, 1, 1-wtm, 1);
			} 
		} else if (score >= beta){	// fail high, research with larger window
			const int RESEARCH_WINDOW = 1500;

			alpha = score;
			beta = score + RESEARCH_WINDOW;

			#if 0
			// don't mind time if not in analyze mode and extra time > 10 seconds
			if (!input_flags.analyze_mode && input_flags.time_left_after > 1000)
				root_info.dont_check_time = 1; // fail low at root, don't check time this ply
			#endif

			score = -Search(-beta, -alpha, new_depth, 1, 1-wtm, 1);
			
			

			if (score >= beta){
				alpha = score;
				beta = INF;
				score = -Search(-beta, -alpha, new_depth, 1, 1-wtm, 1);
				
			
			}

			is_pv = 1;
		}

		nodes = stats.nodes + stats.qnodes  - nodes0;
																				
		current_ply = 0;
		board.Unmake(m, 0, wtm);															
		HashListRemoveLast();																
																						
		if (input_flags.time_left <= 0) return 0; 

		root_info.UpdateMoveInfo(i, score, depth/PLY_DEPTH, nodes);
																																									
		if (is_pv){	
			UpdatePV(0, m);
			root_info.UpdateMoveInfo(i, score, depth/PLY_DEPTH, nodes);
			root_info.MoveToFront(i);
			

			is_pv = 0;

			alpha = score;

			AddHistoryMove(wtm, depth, m);							

			best_move = m;																	

			beta = alpha+1;
			
					
			printf("%i %i %i %i", i, score, depth/PLY_DEPTH, nodes);

		}																					


	}

	if (moves_searched == 0){
		int mate = Check(wtm);

		if (mate) return -MATE;
		else return 0;
	}

	if (alpha > alpha0 && alpha < beta){
		AddHT(alpha, EXACT, depth, best_move);
		AddKiller(best_move, 0);
	} else 
		AddHT(alpha, FAIL_LOW, depth, best_move);

	return alpha;
}





