

typedef unsigned int INDEX;
typedef int square;

#define NEW
#define XX 127
#define T41_INCLUDE

typedef int piece;
typedef int color;

INDEX SqFindKing(square*);
INDEX SqFindFirst(square*, piece);
INDEX SqFindSecond(square*, piece);
INDEX SqFindThird(square*, piece);
INDEX SqFindOne(square*, piece);

#include "tbindex.hpp"

#undef INF
#include "board.h"

#define EGTB_CACHE_SIZE 2*1024*1024


#include <iostream>

using namespace std;


void InitTB(char* path){
	int EGTBMenCount = IInitializeTb(path);
	cout << "Initialized tablebases : " << EGTBMenCount << endl;
	if (EGTBMenCount){
        void *egtb_cache = malloc(EGTB_CACHE_SIZE);
        FTbSetCacheSize(egtb_cache, EGTB_CACHE_SIZE);
    }
}


// returns 1 if succesful and score, else 0
int EGTBProbe(int ply, int wtm, int* score){ 
    int    rgiCounters[10];
    int    iTb;
    color  side;
    int    fInvert;
    square *psqW;
    square *psqB;
    square sqEnP;
    INDEX  ind;

	static int psqw[64], psqb[64];

    // Initialize rgiCounters
	rgiCounters[0] = BitCount(board.wPawns);
	rgiCounters[1] = BitCount(board.wKnights);
	rgiCounters[2] = BitCount(board.wBishops);
	rgiCounters[3] = BitCount(board.wRooks);
	rgiCounters[4] = BitCount(board.wQueens);
	rgiCounters[5] = BitCount(board.bPawns);
	rgiCounters[6] = BitCount(board.bKnights);
	rgiCounters[7] = BitCount(board.bBishops);
	rgiCounters[8] = BitCount(board.bRooks);
	rgiCounters[9] = BitCount(board.bQueens);

    iTb = IDescFindFromCounters (rgiCounters);

    if (iTb == 0)
        return 0; // No such TB

	wtm = 1-wtm; // I use white = 1, black = 0 and it should be white = 0 black = 1

	psqW = psqw;
	psqB = psqb;

	// init psqw, psqb
	for (int i = 0; i < 64; i++){
		if (board.sq[i] <= 0) psqw[i] = 0;
		if (board.sq[i] >= 0) psqb[i] = 0;
		switch(board.sq[i]){
		case  PAWN   : psqw[i] = x_piecePawn; break;
		case  KNIGHT : psqw[i] = x_pieceKnight; break;
		case  BISHOP : psqw[i] = x_pieceBishop; break;
		case  ROOK   : psqw[i] = x_pieceRook; break;
		case  QUEEN  : psqw[i] = x_pieceQueen; break;
		case  KING   : psqw[i] = x_pieceKing; break;

		case -PAWN   : psqb[i] = x_piecePawn; break;
		case -KNIGHT : psqb[i] = x_pieceKnight; break;
		case -BISHOP : psqb[i] = x_pieceBishop; break;
		case -ROOK   : psqb[i] = x_pieceRook; break;
		case -QUEEN  : psqb[i] = x_pieceQueen; break;
		case -KING   : psqb[i] = x_pieceKing; break;
		}
	}

    if (iTb > 0){
		side = wtm;
		fInvert = false;
		psqW = psqw;
		psqB = psqb;
	}else {
		side = 1-wtm;
		fInvert = true;
		iTb = -iTb;
		psqW = psqb;
		psqB = psqw;
	}
    if (!FRegistered(iTb, side))
        return 0;
    
	sqEnP = ((board.ep_square != 0) ? board.ep_square : XX);
    
	ind = PfnIndCalc(iTb, side) (psqW, psqB, sqEnP, fInvert);
    
	int tbValue = L_TbtProbeTable (iTb, side, ind);

	if (tbValue == bev_broken) 
		return 0;

	*score = tbValue*10;

	return 1;

}


INDEX SqFindSecond(square* sq, piece p){
	int counter = 0;
	for (int i = A1; i <= H8; i++, sq++)
		if (*sq == p){
			counter++;
			if (counter == 2) 
				return i;
		}
	return XX;
}

INDEX SqFindThird(square* sq, piece p){
	int counter = 0;
	for (int i = A1; i <= H8; i++, sq++)
		if (*sq == p){
			counter++;
			if (counter == 3) 
				return i;
		}
	return XX;
}

INDEX SqFindOne(square* sq, piece p){
	return SqFindFirst(sq, p);
}
