#include "board.h"
#include "move.h"
#include "prophet.h"
#include <iostream>
#include "tree.h"

using namespace std;

//#define VERIFY_UNMAKEMOVE
//#define VERIFY_MATERIAL
//#define VERIFY_PIECETABLE

//#define VERIFY_PIECECOUNTERS

extern int piecetable[2][7][64];

extern const int piece_v[7];

void Board::Unmake(int move, int ply, int wtm){
	int from = From(move);
	int to = To(move);
	int piece = Piece(move);
	int capture = Captured(move);
	int promote = Promote(move);

	if (capture) 
		piece_counter[1-wtm][capture]++;
	if (promote){
		piece_counter[wtm][promote]--;
		piece_counter[wtm][PAWN]++;
	}

	int ep_move = save_struct[ply].ep_move;
	
	RestoreBoardInfo(ply);

	//cout << "		Unmake move ";
	//PrintMove(move);
	//cout << "(" << from << " " << to << " " << piece << " " << promote << ")" << endl;

	SetClearBoardBit(to, from);

	if (wtm){

		SetClear(white_pieces, to, from);

		if (capture > PAWN)
			black_material++; 
		else if (capture == PAWN)
			black_pawns++;

		material -= piece_v[capture];

		if (promote){
			white_pawns++;
			white_material--;
			material -= piece_v[promote] - piece_v[PAWN];
		}


		if (capture != 0){
			if (ep_move)  // en passant
				board.piecetable_score -= piecetable[BLACK][capture][to-8];
			else 
				board.piecetable_score -= piecetable[BLACK][capture][to];
		}
		board.piecetable_score -= piecetable[WHITE][piece][to] - piecetable[WHITE][piece][from];
		if (piece == KING) {	// castle ?
			if (from == E1){
				if (to == G1)
					board.piecetable_score -= piecetable[WHITE][ROOK][F1] - piecetable[WHITE][ROOK][H1];
				else if (to == C1)
					board.piecetable_score -= piecetable[WHITE][ROOK][D1] - piecetable[WHITE][ROOK][A1];
			}
		}
		if (promote)
			board.piecetable_score -= piecetable[WHITE][promote][to] - piecetable[WHITE][piece][to];



		if (capture && !ep_move){
			SetBoardBit(to);
			Set(black_pieces, to);
		}

		if (!ep_move){
			sq[to] = -capture;
		} else {
			sq[to] = 0;
			sq[to-8] = -PAWN;
		}


		sq[from] = piece;

		SetClear(white_pieces, to, from);

		switch(piece){
		case KNIGHT:
			SetClear(wKnights, to, from);
			break;
		case BISHOP:
			SetClear(wBishops, to, from);
			SetClear(bishops_queens, to, from);
			break;
		case ROOK:
			SetClear(wRooks, to, from);
			SetClear(rooks_queens, to, from);
			break;
		case QUEEN:
			SetClear(wQueens, to, from);
			SetClear(rooks_queens, to, from);
			SetClear(bishops_queens, to, from);
			break;
		case KING:
			SetClear(wKing, to, from);
			white_king_sq = from;

			// castle check
			if (from == E1)
				if (to == G1){
					Assert(sq[F1] == ROOK);
					sq[F1] = 0;
					sq[H1] = ROOK;
					SetClearBoardBit(F1, H1);
					SetClear(white_pieces, F1, H1);
					SetClear(wRooks, F1, H1);
					SetClear(rooks_queens, F1, H1);
				} else if (to == C1){
					Assert(sq[D1] == ROOK);
					sq[D1] = 0;
					sq[A1] = ROOK;
					SetClearBoardBit(D1, A1);
					SetClear(white_pieces, D1, A1);
					SetClear(wRooks, D1, A1);
					SetClear(rooks_queens, D1, A1);
				}
			break;
		case PAWN:
			Set(wPawns, from);
			if (to >= 56){
				switch(promote){
				case QUEEN:
					Clear(wQueens, to);
					Clear(rooks_queens, to);
					Clear(bishops_queens, to);
					break;
				case KNIGHT:
					Clear(wKnights, to);
					break;
				case BISHOP:
					Clear(wBishops, to);
					Clear(bishops_queens, to);
					break;
				case ROOK:
					Clear(wRooks, to);
					Clear(rooks_queens, to);
					break;
				}
			} else {
				Clear(wPawns, to);
			}
			break;
		default:
			Assert(false);
		}

		if (capture)
		switch(capture){
		case KNIGHT:
			Set(bKnights, to);
			break;
		case BISHOP:
			Set(bBishops, to);
			Set(bishops_queens, to);
			break;
		case ROOK:
			Set(bRooks, to);
			Set(rooks_queens, to);
			break;
		case QUEEN:
			Set(bQueens, to);
			Set(rooks_queens, to);
			Set(bishops_queens, to);
			break;
		case PAWN:
			if (ep_move){
				Set(bPawns, to-8);
				SetBoardBit(to-8);
				Set(black_pieces, to-8);
			} else {
				Set(bPawns, to);
			}
			break;
		default:
			Assert(false);
		}
			

	} else {

		SetClear(black_pieces, to, from);

		if (capture > PAWN)
			white_material++; 
		else if (capture == PAWN)
			white_pawns++;

		material += piece_v[capture];

		if (promote){
			black_pawns++;
			black_material--;
			material += piece_v[promote] - piece_v[PAWN];
		}

		if (capture != 0){
			if (ep_move)  // en passant
				board.piecetable_score += piecetable[WHITE][capture][to+8];
			else 
				board.piecetable_score += piecetable[WHITE][capture][to];
		}
		board.piecetable_score += piecetable[BLACK][piece][to] - piecetable[BLACK][piece][from];
		if (piece == KING) {	// castle ?
			if (from == E8){
				if (to == G8)
					board.piecetable_score += piecetable[BLACK][ROOK][F8] - piecetable[BLACK][ROOK][H8];
				else if (to == C8)
					board.piecetable_score += piecetable[BLACK][ROOK][D8] - piecetable[BLACK][ROOK][A8];
			}
		}
		if (promote)
			board.piecetable_score += piecetable[BLACK][promote][to] - piecetable[BLACK][piece][to];


		if (capture && !ep_move){
			SetBoardBit(to);
			Set(white_pieces, to);
		}

		if (!ep_move){
			sq[to] = capture;
		} else {
			sq[to] = 0;
			sq[to+8] = PAWN;
		}


		SetClear(black_pieces, to, from);

		sq[from] = -piece;

		switch(piece){
		case KNIGHT:
			SetClear(bKnights, to, from);
			break;
		case BISHOP:
			SetClear(bBishops, to, from);
			SetClear(bishops_queens, to, from);
			break;
		case ROOK:
			SetClear(bRooks, to, from);
			SetClear(rooks_queens, to, from);
			break;
		case QUEEN:
			SetClear(bQueens, to, from);
			SetClear(rooks_queens, to, from);
			SetClear(bishops_queens, to, from);
			break;
		case KING:
			SetClear(bKing, to, from);
			black_king_sq = from;

			// castle check
			if (from == E8)
				if (to == G8){
					Assert(sq[F8] == -ROOK);
					sq[F8] = 0;
					sq[H8] = -ROOK;
					SetClearBoardBit(F8, H8);
					SetClear(black_pieces, F8, H8);
					SetClear(bRooks, F8, H8);
					SetClear(rooks_queens, F8, H8);
				} else if (to == C8){
					Assert(sq[D8] == -ROOK);
					sq[D8] = 0;
					sq[A8] = -ROOK;
					SetClearBoardBit(D8, A8);
					SetClear(black_pieces, D8, A8);
					SetClear(bRooks, D8, A8);
					SetClear(rooks_queens, D8, A8);
				}
			break;
		case PAWN:
			Set(bPawns, from);
			if (to <= 7){
				switch(promote){
				case QUEEN:
					Clear(bQueens, to);
					Clear(rooks_queens, to);
					Clear(bishops_queens, to);
					break;
				case KNIGHT:
					Clear(bKnights, to);
					break;
				case BISHOP:
					Clear(bBishops, to);
					Clear(bishops_queens, to);
					break;
				case ROOK:
					Clear(bRooks, to);
					Clear(rooks_queens, to);
					break;
				}
			} else {
				Clear(bPawns, to);
			}
			break;
		default:
			Assert(false);
		}

		if (capture)
		switch(capture){
		case KNIGHT:
			Set(wKnights, to);
			break;
		case BISHOP:
			Set(wBishops, to);
			Set(bishops_queens, to);
			break;
		case ROOK:
			Set(wRooks, to);
			Set(rooks_queens, to);
			break;
		case QUEEN:
			Set(wQueens, to);
			Set(rooks_queens, to);
			Set(bishops_queens, to);
			break;
		case PAWN:
			if (ep_move){
				Set(wPawns, to+8);
				SetBoardBit(to+8);
				Set(white_pieces, to+8);
			} else {
				Set(wPawns, to);
			}
			break;
		default:
			Assert(false);
		}

	}


	#ifdef VERIFY_UNMAKEMOVE
		Verify("UnmakeMove");
	#endif

	#ifdef VERIFY_MATERIAL
		VerifyMaterial();
	#endif

	#ifdef VERIFY_PIECETABLE
		Assert(board.piecetable_score == EvaluateStaticPosition());
	#endif

	#ifdef VERIFY_PIECECOUNTERS
		board.VerifyPieceCounter();
	#endif
}
